def change_global_param(self, param_name, new_value):
    """
    Used to change a param in global params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.global_params:
        self.global_params[param_name] = new_value
        self.save_new_param_json()
        return True
    else:
        return False


def change_dataset_param(self, param_name, new_value):
    """
    Used to change a param in dataset params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.dataset_params:
        self.dataset_params[param_name] = new_value
        self.global_params["need_recalc_dataset_params"] = True
        self.global_params["need_recalc_box_params"] = True
        self.global_params["need_recalc_eigen_params"] = True
        # self.save_new_param_json()
        return True
    else:
        return False


def change_filter_param(self, param_name, new_value):
    """
    Used to change a param in filter params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.filter_params:
        self.filter_params[param_name] = new_value
        self.global_params["need_recalc_filter_params"] = True
        self.global_params["need_recalc_box_params"] = True

        self.global_params["need_recalc_eigen_params"] = True
        # self.save_new_param_json()
        return True
    else:
        return False


def change_box_param(self, param_name, new_value):
    """
    Used to change a param in box params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.box_params:
        # if param_name == "total_num_spatial_boxes":
        #     assert (int(new_value**.5))**2 == new_value, "Please make sure Number of Spatial Boxes is a square number"
        self.box_params[param_name] = new_value
        self.global_params["need_recalc_box_params"] = True
        self.global_params["need_recalc_eigen_params"] = True
        self.global_params["need_recalc_roi_extraction_params"] = True
        # self.save_new_param_json()
        return True
    else:
        return False


def change_eigen_param(self, param_name, new_value):
    """
    Used to change a param in eigen params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.eigen_params:
        self.eigen_params[param_name] = new_value
        self.global_params["need_recalc_eigen_params"] = True
        # self.save_new_param_json()
        return True
    else:
        return False


def change_roi_extraction_param(self, param_name, new_value):
    """
    Used to change a param in roi_extraction params area.
    Parameters
    ----------
    param_name : str
        name of parameter to change
    new_value : ?
        the new value to the parameter

    Returns
    -------
    True if successful
    """
    if param_name in self.roi_extraction_params:
        self.roi_extraction_params[param_name] = new_value
        self.global_params["need_recalc_roi_extraction_params"] = True
        # self.save_new_param_json()
        return True
    else:
        return False



import json
import logging

from cidan.LSSC.functions.pickle_funcs import *

logger1 = logging.getLogger("cidan.DataHandler")


def load_param_json(self):
    """
    Loads the parameter json file and saves it to all the parameter values
    """

    try:
        with open(self.param_path, "r") as f:
            all_params = json.loads(f.read())
        self.global_params = all_params["global_params"]
        self.dataset_params = all_params["dataset_params"]
        self.filter_params = all_params["filter_params"]
        self.box_params = all_params["box_params"]
        self.box_params_processed = all_params["box_params"].copy()
        self.eigen_params = all_params["eigen_params"]
        self.time_trace_params = all_params["time_trace_params"]
        self.roi_extraction_params = all_params["roi_extraction_params"]
        self.trials_loaded = self.dataset_params["trials_loaded"]
        self.trials_all = self.dataset_params["trials_all"]
        return True
    except KeyError:
        raise KeyError("Please Choose a valid parameter file")
    except FileNotFoundError:
        raise FileNotFoundError("Can't find parameter file")
    except NameError:
        raise FileNotFoundError("Can't find parameter file")


def save_new_param_json(self):
    """
    Saves the parameters to the parameter file
    """
    try:
        with open(self.param_path, "w") as f:
            all_params = {
                "global_params": self.global_params,
                "dataset_params": self.dataset_params,
                "filter_params": self.filter_params,
                "box_params": self.box_params_processed,
                "eigen_params": self.eigen_params,
                "roi_extraction_params": self.roi_extraction_params,
                "time_trace_params": self.time_trace_params
            }
            f.truncate(0)
            f.write(json.dumps(all_params))
    except:
        raise FileNotFoundError("Error saving parameters, please restart software")
    pass


def create_new_save_dir(self):
    """
    Creates a new save directory
    """
    try:
        if not os.path.isdir(self.save_dir_path):
            os.mkdir(self.save_dir_path)
        eigen_vectors_folder_path = os.path.join(self.save_dir_path,
                                                 "eigen_vectors/")

        if not os.path.isdir(eigen_vectors_folder_path):
            os.mkdir(eigen_vectors_folder_path)
        embedding_images_path = os.path.join(self.save_dir_path,
                                             "embedding_norm_images/")
        if not os.path.isdir(embedding_images_path):
            os.mkdir(embedding_images_path)
        eigen_vectors_folder_path = os.path.join(self.save_dir_path,
                                                 "temp_files/")

        if not os.path.isdir(eigen_vectors_folder_path):
            os.mkdir(eigen_vectors_folder_path)
        return True
    except:
        raise FileNotFoundError("Couldn't create folder please try again")



import os
import warnings
from typing import Tuple, List

import matplotlib.pyplot as plt
import numpy as np
import tifffile
from PIL import Image
from dask import delayed
from scipy import ndimage, sparse
# from IPython.display import display, Image
from sklearn.decomposition import PCA

from cidan.LSSC.functions.embeddings import calcDInv
from cidan.LSSC.functions.widefield_functions import *


def load_tif_stack(path, mask_flat=None, convert_to_32=True):
    """
    This function reads a tiff stack file
    Parameters
    ----------
    path The path to a single tif stack or to a directory of tiff stacks


    Returns
    -------
    a 3D numpy array with the tiff files together or a zarr array if not loading into memory

    """

    size = [0, 0]

    if os.path.isdir(path):
        volumes = []
        paths = path if type(path) == list else sorted(os.listdir(path))
        # if trial_split:
        #     paths = paths[trial_num * trial_split_length:(
        #                                                          trial_num + 1) * trial_split_length if (
        #                                                                                                         trial_num + 1) * trial_split_length < len(
        #         paths) else len(paths)]
        for num, x in enumerate(paths):
            file_path = x if type(path) == list else os.path.join(path, x)
            image = tifffile.imread(file_path)
            size = [image.shape[0], image.shape[1]]
            if len(image.shape) == 2:
                image = image.reshape((1, image.shape[0], image.shape[1]))
            volumes.append(image)


        image = np.vstack(volumes)
        del volumes
    elif os.path.isfile(path):

        image = tifffile.imread(path)
        size = [image.shape[1], image.shape[2]]
        if len(image.shape) == 2:
            image = image.reshape((1, image.shape[0], image.shape[1]))


    else:
        raise Exception("Invalid Inputs ")

    # print(np.mean(image[0]),np.max(image[0]),np.min(image[0]),np.median(image[0]))

    if convert_to_32:
        image = image.astype(np.float32)
        if np.isclose(np.mean(image[0]), 2 ** 15, 0, 4000):
            image = image - 2 ** 15
            image[image < 0] = 0
    if mask_flat is not None:
        image = image.transpose((1, 2, 0))
        try:
            image[mask_flat.reshape((image.shape[0], image.shape[1])) == 0] = 0
        except:
            print("Error mask is the wrong size")
        image = image.transpose((2, 0, 1))
    return image

def load_tif_stack_trial(path):
    """
        This function loads only a specific trial
        Parameters
        ----------
        path The path to zarr file of a single tiff stack, a folder with trials, or a folder with tiff images


        Returns
        -------
        a 3D numpy array

        """
    pass
# TODO Implement this

def reshape_to_2d_over_time(volume):
    """
    Takes a 3d numpy volume with dim 1 and 2 being x, y and dim 0 being time
    and return array compressing dim 1 and 2 into 1 dimension

    Parameters
    ----------
    volume input 3d numpy volume

    Returns
    -------
    a 2d volume with the 0 dim being pixel number and 1 being time value

    """
    return np.transpose(np.reshape(volume, (volume.shape[0], -1), order="C"))


def save_image(image: np.ndarray, path):
    """
    Function to save image
    Parameters
    ----------
    image
    name
    directory


    Returns
    -------
    None
    """
    # pass
    percent = 99
    image_temp = image.copy()
    image[image < 0] = 0
    image = (((image - np.percentile(image, 1)) / (
            np.percentile(image, percent) - np.percentile(
        image, 1))))
    if np.percentile(image, 30) > .25:
        image = (
            ((image_temp - np.percentile(image_temp, 10)) / (
                    np.percentile(image_temp, percent) - np.percentile(
                image_temp, 10))))

    image[image > 1] = 1
    image = image * 255
    image[image < 0] = 0

    combined_image = np.dstack([np.zeros_like(image), image,
                                np.zeros_like(image)])
    final_image = combined_image.astype(np.uint8)
    plt.imsave(path, final_image, vmin=0, vmax=255)


def save_image_multiple(volume: np.ndarray, name: str, directory: str, shape: tuple,
               number_save: int):
    """
    Function to save images from a 3d volume
    Parameters
    ----------
    volume
    name
    directory
    shape shape to reshape volume into
    number_save number of slices to take and save

    Returns
    -------
    None
    """
    # pass
    for x in range(number_save):
        fig, ax = plt.subplots()
        imgplot = ax.imshow(np.reshape(volume,
                                       shape)[shape[0] // number_save * x])
        fig.savefig(os.path.join(directory, name + "_" + str(x)))

        img = Image.fromarray(
            np.reshape(volume,
                       shape)[shape[0] / number_save * x] * 255).convert('L')
        img.save(
            os.path.join(directory, name + "_" + str(x)))


def filter_stack(*, stack: np.ndarray, median_filter: bool,
                 median_filter_size: Tuple[int, int, int],
                 z_score: bool, hist_eq=False, localSpatialDenoising=False):
    stack = stack.astype(np.float32)
    if np.isclose(np.mean(stack[0]), 2 ** 15, 0, 4000):
        stack = stack - 2 ** 15
        stack[stack < 0] = 0
    if localSpatialDenoising:
        stack = applyLocalSpatialDenoising(stack)
    if z_score and not hist_eq:
        stack_t = np.transpose(stack, (1, 2, 0))
        shape = (stack.shape[1], stack.shape[2], 1)
        std = np.std(stack_t, axis=2).reshape(shape)+1E-6
        mean = np.mean(stack_t, axis=2).reshape(shape)
        stack_t = (stack_t - mean) / std
        stack = np.transpose(stack_t, (2, 0, 1))
    if hist_eq:
        stack_t = np.transpose(stack, (1, 2, 0))
        quant_5 = np.percentile(stack_t, 5, axis=2)
        quant_95 = np.percentile(stack_t, 95, axis=2)
        quant_5_filtered = ndimage.filters.gaussian_filter(quant_5, 2)
        quant_95_filtered = ndimage.filters.gaussian_filter(quant_95, 2)
        stack_equalized = np.divide(
            np.subtract(stack_t, quant_5_filtered[..., np.newaxis]), (
                    quant_95_filtered - quant_5_filtered + .00000001)[..., np.newaxis])

        stack_equalized = np.nan_to_num(stack_equalized)

        stack_equalized[stack_equalized > 1] = 1
        stack_equalized[stack_equalized < 0] = 0
        stack_equalized_squared = np.power(stack_equalized, 2)
        stack_equalized_filtered = ndimage.median_filter(stack_equalized_squared,
                                                         (1, 1, 3))
        stack = np.transpose(stack_equalized_filtered, (2, 0, 1))

    if median_filter:
        stack = ndimage.median_filter(stack, median_filter_size)
    return stack
def slice_crop_stack( image,slice_stack: bool,
                          slice_every, slice_start: int, crop_stack: bool,
                          crop_x: List[int], crop_y: List[int]):
    if slice_stack:
        image = image[slice_start::slice_every]
    if crop_stack:
        image = image[:, crop_x[0]:crop_x[1], crop_y[0]:crop_y[1]]
    return image
def auto_crop(dataset_list):
    # takes in raw dataset or list of datasets and outputs suggested crops
    suggested_crops = [0 for _ in dataset_list]
    for num,dataset in enumerate(dataset_list):
        crop_y_bools = (dataset <= dataset.min()).all(1).any(0)
        y_iter_1 = 0
        while True:
            if crop_y_bools[y_iter_1]:
                y_iter_1 += 1
            else:
                break
        y_iter_2 = 0
        while True:
            if crop_y_bools[-(y_iter_2 + 1)]:
                y_iter_2 += 1
            else:
                break

        crop_x_bools = (dataset <= dataset.min()).all(2).any(0)
        x_iter_1 = 0
        while True:
            if crop_x_bools[x_iter_1]:
                x_iter_1 += 1
            else:
                break
        x_iter_2 = 0
        while True:
            if crop_x_bools[-(x_iter_2 + 1)]:
                x_iter_2 += 1
            else:
                break
        suggested_crops[num] = [[x_iter_1, x_iter_2], [y_iter_1, y_iter_2]]
    crop_size = [[max([x[0][0] for x in suggested_crops]),
             -max([x[0][1] for x in suggested_crops])],
            [max([x[1][0] for x in suggested_crops]),
             -max([x[1][1] for x in suggested_crops])]]
    crop_final = [[0,0],[0,0]]
    crop_final[0][0] = 0 + \
                                       crop_size[0][0]
    crop_final[0][1] = dataset_list[0].shape[1] + \
                                       crop_size[0][1]
    crop_final[1][0] = 0 + \
                                       crop_size[1][0]
    crop_final[1][1] = dataset_list[0].shape[2]  + \
                                       crop_size[1][1]
    return crop_final
def pixel_num_to_2d_cord(pixel_list, volume_shape: Tuple[int, int]):
    """
    Takes a list of pixels and converts it to a list of cords for each pixel

    Parameters
    ----------
    pixel_list : np.ndarray
        list of pixel numbers
    volume_shape : tuple
        shape of the original dataset [x,y]

    Returns
    -------
    a list of tuples where each is the cordinates for each pixel

    """

    def convert_one(num):
        return num % volume_shape[1], num // volume_shape[1]

    return np.apply_along_axis(convert_one, axis=0, arr=pixel_list)


def cord_2d_to_pixel_num(cord_list, volume_shape: Tuple[int, int]):
    """
    Takes a list of cords and converts it to a list of pixel number for each pixel

    Parameters
    ----------
    pixel_list : np.ndarray
        list of cords
    volume_shape : tuple
        shape of the original dataset [x,y]

    Returns
    -------
    a list of numbers for pixel # for each pixel

    """

    def convert_one(num):
        return num[0] * volume_shape[1] + num[1]

    return np.apply_along_axis(convert_one, axis=0, arr=cord_list)

@delayed
def join_data_list(data_list):
    """
    Joins a list of trials with the same dimensions
    Parameters
    ----------
    data_list
        Data to stack

    Returns
    -------
    List of data stacked over 1st dimension
    """
    return np.vstack(data_list)


@delayed
def saveTempImage(data, save_dir, spatial_box_num):
    # print(save_dir)
    temp_dir = os.path.join(save_dir, "temp_images")
    # print(data[60, :, :].shape)
    img = Image.fromarray(data[3, :, :] * 255).convert('L')
    image_path = os.path.join(temp_dir, "embedding_norm_image_box_{}.png".format(
        spatial_box_num))
    img.save(image_path)
    return data


def applyPCA(data, pca_threshold, mask_flat=None):
    data_shape = data.shape
    from cidan.LSSC.functions.roi_extraction import elbow_threshold
    if mask_flat is not None:
        data = mask_data_3d(data, mask_flat)
    else:
        data = reshape_to_2d_over_time(data)
    pca = PCA()
    pca.fit(data.transpose())
    threshold = elbow_threshold(pca.singular_values_[:100],
                                np.arange(0, len(pca.singular_values_[:100]), 1),
                                half=False)
    print(threshold, pca.singular_values_)
    filtered_pca_components = pca.components_[pca.singular_values_ > threshold]
    if mask_flat is not None:
        filtered_pca_components = mask_to_data_2d(filtered_pca_components.transpose(),
                                                  mask_flat).transpose()
    return filtered_pca_components.reshape((-1, data_shape[1], data_shape[2]))


def applyLocalSpatialDenoising(data):
    data = data.transpose((1, 2, 0))
    shape = data.shape

    affinity_self = np.sum(np.power(data, 2), axis=2) + .0000001
    affinity_up = np.divide(
        np.sum(np.multiply(data[0:-1, :, :], data[1:, :, :]), axis=2),
        np.multiply(np.power(affinity_self[0:-1, :], .5),
                    np.power(affinity_self[1:, :], .5)))
    affinity_right = np.divide(
        np.sum(np.multiply(data[:, :-1, :], data[:, 1:, :]), axis=2),
        np.multiply(np.power(affinity_self[:, :-1], .5),
                    np.power(affinity_self[:, 1:], .5)))
    affinity_up_right = np.divide(
        np.sum(np.multiply(data[:-1, 1:, :], data[1:, :-1, :]), axis=2),
        np.multiply(np.power(affinity_self[:-1, 1:], .5),
                    np.power(affinity_self[1:, :-1], .5)))
    affinity_down_right = np.divide(
        np.sum(np.multiply(data[:-1, :-1, :], data[1:, 1:, :]), axis=2),
        np.multiply(np.power(affinity_self[:-1, :-1], .5),
                    np.power(affinity_self[1:, 1:], .5)))
    indices_self = np.arange(0, shape[0] * shape[1], 1)
    indices_right_x = np.arange(0, shape[0] * shape[1], 1)

    indices_right_x = indices_right_x[indices_right_x % shape[1] != shape[1] - 1]
    indices_right_y = np.arange(0, shape[0] * shape[1], 1)
    indices_right_y = indices_right_y[indices_right_y % shape[1] != 0]
    indices_up_x = np.arange(shape[1], shape[0] * shape[1], 1)
    indices_up_y = np.arange(0, shape[0] * shape[1] - shape[1], 1)
    indices_up_right_x = np.arange(shape[1], shape[0] * shape[1], 1)
    indices_up_right_x = indices_up_right_x[
        indices_up_right_x % shape[1] != shape[1] - 1]
    indices_up_right_y = np.arange(0, shape[0] * shape[1] - shape[1], 1)
    indices_up_right_y = indices_up_right_y[indices_up_right_y % shape[1] != 0]
    indices_down_right_x = np.arange(0, shape[0] * shape[1] - shape[1], 1)
    indices_down_right_x = indices_down_right_x[
        indices_down_right_x % shape[1] != shape[1] - 1]
    indices_down_right_y = np.arange(shape[1], shape[0] * shape[1], 1)
    indices_down_right_y = indices_down_right_y[indices_down_right_y % shape[1] != 0]

    K = sparse.csr_matrix((np.concatenate(
        [np.reshape(affinity_up, (-1)), np.reshape(affinity_right, (-1)),
         np.reshape(affinity_up_right, (-1)), np.reshape(affinity_down_right, (-1))]),
                           (np.concatenate([indices_up_x, indices_right_x,
                                            indices_up_right_x,
                                            indices_down_right_x]),
                            np.concatenate([indices_up_y, indices_right_y,
                                            indices_up_right_y,
                                            indices_down_right_y]))))
    K = K + K.transpose()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        K.setdiag(1)
    D_inv, D_diag = calcDInv(K)
    P = D_inv.dot(K)
    temp_flat = reshape_to_2d_over_time(data.transpose(2, 0, 1))
    return P.dot(temp_flat).reshape((shape[0], shape[1], shape[2])).transpose((2, 0, 1))



import logging
import random

import numpy as np
import pyqtgraph as pg
import qdarkstyle
from qtpy.QtWidgets import *
from skimage import measure

from cidan.GUI.Data_Interaction import DataHandler
from cidan.GUI.Data_Interaction.ROIExtractionThread import ROIExtractionThread
from cidan.GUI.ImageView.ROIImageViewModule import ROIImageViewModule
from cidan.GUI.Inputs.ColorInput import ColorInput
from cidan.GUI.Inputs.OptionInput import OptionInput
from cidan.GUI.Inputs.StringInput import StringInput
from cidan.GUI.ListWidgets.ClassListModule import ClassListModule
from cidan.GUI.ListWidgets.ROIListModule import ROIListModule
from cidan.GUI.ListWidgets.TrialListWidget import TrialListWidget
from cidan.GUI.Tabs.ROIExtractionTab import ROIExtractionTab
from cidan.GUI.Tabs.Tab import Tab
from cidan.TimeTrace.signal_to_noise import trial_PSNR

logger1 = logging.getLogger("cidan.ROIExtractionTab")


class ClassificationTab(Tab):
    """Class controlling the Classification tab, inherits from Tab


        Attributes
        ----------
        main_widget : MainWidget
            A reference to the main widget
        data_handler : DataHandler
            A reference to the main DataHandler of MainWidget
        time_plot : pg.PlotWidget
            the plot for the time traces
        roi_list_module : ROIListModule
            The module the controlls the list of ROIs
        thread : ROIExtractionThread
            The thread that runs the roi extraction process

        """

    def __init__(self, main_widget):

        self.main_widget = main_widget

        self.image_view = ROIImageViewModule(main_widget, self, False,
                                             display_class_option=True)
        self.image_view.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.roi_list_module = ROIListModule(main_widget.data_handler, self,
                                             select_multiple=False, display_time=False,
                                             class_version=True)
        self.image_view.outlines = False
        self.image_view.current_forground = "Class"
        # self.tab_selector_roi.setStyleSheet("QTabWidget {font-size: 20px;}")

        display_setting_layout = QVBoxLayout()
        display_setting_layout.addWidget(QLabel("Display Settings"))
        self.display_settings = self.image_view.createSettings()
        display_setting_layout.addWidget(self.display_settings)
        roi_classification_layout = QVBoxLayout()

        self.button_layout_1 = QHBoxLayout()
        self.button_layout_2 = QHBoxLayout()

        self.stats_layout = QVBoxLayout()
        self.roi_label = QLabel(
            "ROI #{}:".format(self.roi_list_module.current_selected_roi))
        roi_classification_layout.addWidget(self.roi_label)
        stats_layout_h1 = QHBoxLayout()
        stats_layout_h2 = QHBoxLayout()
        self.stat_labels = {"Area": QLabel("Area: None"),
                            "Eccentricity": QLabel("Eccentricity: None"),
                            "Extent": QLabel("Extent: None"),
                            "PSNR": QLabel("PSNR: None")}
        stats_layout_h1.addWidget(self.stat_labels["Area"])
        stats_layout_h1.addWidget(self.stat_labels["Eccentricity"])

        stats_layout_h2.addWidget(self.stat_labels["Extent"])
        stats_layout_h2.addWidget(self.stat_labels["PSNR"])
        # stats_layout_h1.addWidget(QLabel(""))
        # stats_layout_h2.addWidget(QLabel(""))
        self.stats_layout.addLayout(stats_layout_h1)
        self.stats_layout.addLayout(stats_layout_h2)
        roi_classification_layout.addLayout(self.stats_layout)
        roi_classification_layout.addLayout(self.button_layout_1)
        roi_classification_layout.addLayout(self.button_layout_2)

        self.roi_class_dict_tabs = QTabWidget()
        self.roi_class_dict_tabs.addTab(self.roi_list_module, "ROI List")

        self.class_list_module = ClassListModule(classifier_tab=self,
                                                 data_handler=self.data_handler, )
        self.create_class_button = QPushButton("Create New Class")
        self.create_class_button.clicked.connect(lambda x: self.create_class())
        class_layout = QVBoxLayout()
        class_layout.addWidget(self.class_list_module)
        class_layout.addWidget(self.create_class_button)
        class_layout_widget = QWidget()
        class_layout_widget.setLayout(class_layout)

        self.roi_class_dict_tabs.addTab(class_layout_widget, "Classes")

        # Tab selector for the time trace window
        tab_selector_time_trace = QTabWidget()
        # tab_selector_time_trace.setStyleSheet("QTabWidget {font-size: 20px;}")
        tab_selector_time_trace.setMaximumHeight(220)
        # plot of time traces
        self.time_plot = pg.PlotWidget()
        self.time_plot.getPlotItem().getViewBox().setMouseEnabled(True, False)
        self.time_plot.showGrid(x=True, y=True, alpha=0.3)

        tab_selector_time_trace.addTab(self.time_plot, "Time Trace Plot")
        # Setting window for time traces
        time_trace_settings = QWidget()
        time_trace_settings_layout = QVBoxLayout()
        time_trace_settings_layout.setContentsMargins(0, 0, 0, 0)
        time_trace_settings.setLayout(time_trace_settings_layout)
        self.time_trace_type = OptionInput("Time Trace Type", "",
                                           lambda x, y: self.deselectRoiTime(),
                                           default_index=0,
                                           tool_tip="Select way to calculate time trace,"
                                                    " \ncheck github for more details",
                                           val_list=list(
                                               self.data_handler.time_trace_possibilities_functions.keys()))
        time_trace_settings_layout.addWidget(self.time_trace_type,
                                             stretch=1)
        # A list widget to select what trials to calculate/display time traces for
        self._time_trace_trial_select_list = TrialListWidget(False)
        self._time_trace_trial_select_list.setMinimumHeight(115)
        self._time_trace_trial_select_list.set_items_from_list(
            self.data_handler.trials_all,
            self.data_handler.trials_loaded_time_trace_indices)
        if not len(self.data_handler.trials_loaded) == 1 and not \
                self.data_handler.dataset_params["single_file_mode"] and not \
                self.data_handler.dataset_params["trial_split"]:
            time_trace_settings_layout.addWidget(self._time_trace_trial_select_list,
                                                 stretch=5)
        time_trace_update_button = QPushButton("Update Time Traces")
        time_trace_settings_layout.addWidget(time_trace_update_button)
        time_trace_update_button.clicked.connect(
            lambda x: self.update_time_traces())
        tab_selector_time_trace.addTab(time_trace_settings, "Time Trace Settings")
        self.roi_list_module.keyPressEvent = self.keyPressEvent
        self.roi_class_dict_tabs.keyPressEvent = self.keyPressEvent
        self.updateTab()

        super().__init__("Classification",
                         column_1=[self.display_settings, roi_classification_layout,
                                   self.roi_class_dict_tabs],
                         column_2=[self.image_view, tab_selector_time_trace],
                         column_2_display=True, horiz_moveable=True)
    def deselectRoiTime(self):
        ROIExtractionTab.deselectRoiTime(self)
    def update_time_traces(self):
        ROIExtractionTab.update_time_traces(self)

    def keyPressEvent(self, event):
        super(ClassificationTab, self).keyPressEvent(event)
        print(event.key())
        # add_thing = 49 # key number for key "1"
        # for x in range(0,9,1):
        #     if event.key() == add_thing+x and len(self.data_handler.classes.keys())>x:
        #
        #         self.assign_roi_to_class(list(self.data_handler.classes.keys())[x])
        # if event.key()== add_thing-1 and len(self.data_handler.classes.keys())>=10:# for key 0 which will assign class 10
        #     self.assign_roi_to_class(list(self.data_handler.classes.keys())[9])

    @property
    def data_handler(self):
        return self.main_widget.data_handler

    def set_class_buttons(self):
        def button_function(class_id):
            return lambda x: self.assign_roi_to_class(class_id)

        for i in reversed(range(self.button_layout_1.count())):
            self.button_layout_1.itemAt(i).widget().setParent(None)
        for i in reversed(range(self.button_layout_2.count())):
            self.button_layout_2.itemAt(i).widget().setParent(None)
        ids = list(self.data_handler.classes.keys())
        for num, id in enumerate(ids):
            button = QPushButton(self.data_handler.classes[id]["name"])

            button.clicked.connect(button_function(id))
            if num < 4:
                self.button_layout_1.addWidget(button)
            else:
                self.button_layout_2.addWidget(button)

        self.class_list_module.set_list_items(self.data_handler.classes)
        # button = QPushButton("Delete ROI")
        # button.clicked.connect(lambda x: self.delete_roi())

    def assign_roi_to_class(self, class_name):
        if self.roi_list_module.current_selected_roi is not None:
            try:
                self.data_handler.assign_roi_class(
                    self.roi_list_module.current_selected_roi, new_class=class_name,
                    input_key=False)
            except AttributeError:
                pass
            self.updateTab()
            try:
                next_roi = self.main_widget.data_handler.roi_index_backward[self.roi_list_module.current_selected_roi+1]
                self.image_view.zoomRoi(self.data_handler.rois_dict[next_roi]["index"], input_key=False)
                self.roi_list_module.set_current_select(next_roi)
            except IndexError:
                pass

    def create_class(self):
        if len(self.data_handler.classes.keys()) >= 8:
            msg = QMessageBox()
            msg.setStyleSheet(qdarkstyle.load_stylesheet())
            msg.setWindowTitle("Notice")

            msg.setText(
                "The maximum number of allowed classes is 8")
            msg.setIcon(QMessageBox.Information)
            x = msg.exec_()
            return
        id = len(self.data_handler.classes.keys())

        while id in self.data_handler.classes.keys():
            id = random.randint(a=1000, b=100000)
        self.data_handler.classes[str(id)] = {"name": "default name", "rois": [],
                                              "color": self.data_handler.color_list[
                                                  id % len(
                                                      self.data_handler.color_list)],
                                              "editable": True}
        self.edit_class(str(id))

    def delete_class(self, class_id):

        if not self.data_handler.classes[class_id]["editable"]:
            msg = QMessageBox()
            msg.setStyleSheet(qdarkstyle.load_stylesheet())
            msg.setWindowTitle("Warning")

            msg.setText("You can't edit this class")
            msg.setIcon(QMessageBox.Warning)
            x = msg.exec_()
            return
        del self.data_handler.classes[class_id]
        self.updateTab()

    def edit_class(self, class_id):
        class_name = self.data_handler.classes[class_id]["name"]
        if not self.data_handler.classes[class_id]["editable"]:
            msg = QMessageBox()
            msg.setStyleSheet(qdarkstyle.load_stylesheet())
            msg.setWindowTitle("Warning")

            msg.setText("You can't edit this class")
            msg.setIcon(QMessageBox.Warning)
            x = msg.exec_()
            return
        dialog = QDialog()
        dialog.layout = QVBoxLayout()
        self.name_input = StringInput(default_val=class_name,
                                      display_name="Class Name:",
                                      on_change_function=None,
                                      program_name="Class Name:",
                                      tool_tip="The name of the class")
        dialog.layout.addWidget(self.name_input)
        self.color_input = ColorInput(
            default_val=self.data_handler.classes[class_id]["color"],
            display_name="Color:", program_name="Color", on_change_function=None,
            tool_tip="")
        dialog.layout.addWidget(self.color_input)
        style = str("""
                    QVBoxLayout {
                
                        border:%dpx;
                    }


                    """ % (2))
        dialog.setStyleSheet(qdarkstyle.load_stylesheet() + style)

        def finish_func(delete=False):
            dialog.close()
            self.data_handler.classes[class_id][
                "color"] = self.color_input.current_state()
            self.data_handler.classes[class_id][
                "name"] = self.name_input.current_state()
            if delete:
                self.delete_class(class_id)
            else:
                self.updateTab()

        dialog.setWindowTitle("Edit Class:")
        delete_button = QPushButton("Delete Class")
        finish_button = QPushButton("Done")
        delete_button.clicked.connect(lambda x: finish_func(delete=True))
        finish_button.clicked.connect(lambda x: finish_func())
        horiz_button_layout = QHBoxLayout()
        horiz_button_layout.addWidget(delete_button)
        horiz_button_layout.addWidget(finish_button)

        dialog.layout.addLayout(horiz_button_layout)

        dialog.setLayout(dialog.layout)
        dialog.show()

    def selectRoiTime(self, num):

        try:
            color_roi = self.main_widget.data_handler.color_list[
                (num) % len(self.main_widget.data_handler.color_list)]

            if (self.roi_list_module.roi_time_check_list[num]):
                if self.data_handler.roi_time_trace_need_update[num]:
                    self.main_widget.console.updateText(
                        "Some time traces are out of date, please recalculate",
                        warning=True)
                pen = pg.mkPen(color=color_roi, width=3)
                time_trace = self.main_widget.data_handler.get_time_trace(num,
                                                                          trace_type=self.time_trace_type.current_state())
                if type(time_trace) == bool:
                    return
                self.time_plot.plot(time_trace,
                                    pen=pen)
                self.time_plot.enableAutoRange(axis=0)
        except AttributeError:
            pass

    def update_current_roi_selected(self):
        self.roi_label.setText(
            "ROI #{}:".format(self.roi_list_module.current_selected_roi))
        self.time_plot.clear()
        if self.roi_list_module.current_selected_roi is None:
            self.stat_labels["Area"].setText("Area: None")
            self.stat_labels["Eccentricity"].setText("Eccentricity: None")

            self.stat_labels["Extent"].setText("Extent: None")
            self.stat_labels["PSNR"].setText("PSNR: None")
            return
        try:

            roi_index = self.roi_list_module.current_selected_roi
            self.selectRoiTime(roi_index)
            self.image_view.zoomRoi(roi_index, input_key=False, update=False)
            roi_pix = self.data_handler.rois[roi_index]
            original_zeros = np.zeros(self.data_handler.shape, dtype=int).reshape(
                (-1, 1))
            original_zeros[roi_pix] = 1
            original_zeros = original_zeros.reshape(self.data_handler.shape)
            props = measure.regionprops(original_zeros)[0]
            # for key in props:
            #     try:
            #         print(key, float(props[key]))
            #     except:
            #         pass
            extent = len(roi_pix) / ((self.data_handler.roi_max_cord_list[roi_index][
                                          0] -
                                      self.data_handler.roi_min_cord_list[roi_index][
                                          0]) * (self.data_handler.roi_max_cord_list[
                                                     roi_index][1] -
                                                 self.data_handler.roi_min_cord_list[
                                                     roi_index][1]))
            time_trace = self.data_handler.get_time_trace(roi_index)
            psnr = trial_PSNR(time_trace.reshape((1, -1)))[0]
            self.stat_labels["Area"].setText("Area: {}".format(len(roi_pix)))
            self.stat_labels["Eccentricity"].setText(
                "Eccentricity: {:.2f}".format(props["eccentricity"]))
            self.stat_labels["Extent"].setText("Extent: {:.2f}".format(extent))
            self.stat_labels["PSNR"].setText("PSNR: {:.2f}".format(psnr))



        except:
            # print("SOmething happened")
            pass

    def keyPressAction(self, event):

        if event.key() == 16777234:
            self.roi_list_module.select_roi_next(False)
            event.accept()
        if event.key() == 16777236:
            self.roi_list_module.select_roi_next(True)
            event.accept()

        return event.isAccepted()

    def updateTab(self, gen_display=False):
        if (self.main_widget.checkThreadRunning()):

            self._time_trace_trial_select_list.set_items_from_list(
                self.data_handler.trials_all,
                self.data_handler.trials_loaded_time_trace_indices)
            if self.data_handler.rois_loaded:
                if gen_display:
                    self.data_handler.gen_class_display_variables()
                self.roi_list_module.set_list_items(
                    self.main_widget.data_handler.rois_dict)
            self.set_class_buttons()
            self.roi_label.setText(
                "ROI #{}:".format(self.roi_list_module.current_selected_roi))
            self.image_view.reset_view()



import json
import logging
from os import listdir
from os.path import isfile, join

import matplotlib.pyplot as plt
import numpy as np
from scipy.io import savemat
from skimage import measure
from skimage.measure import find_contours

from cidan.LSSC.SpatialBox import SpatialBox
from cidan.LSSC.functions.data_manipulation import reshape_to_2d_over_time
from cidan.LSSC.functions.pickle_funcs import *
from cidan.LSSC.functions.spatial_footprint import classify_components_ep

logger1 = logging.getLogger("cidan.DataHandler")


def calculate_statistics(self):
    statistics = [False for x in range(len(self.rois))]
    A = np.zeros([self.edge_roi_image_flat.shape[0], len(self.rois)], dtype=int)
    for num, roi in enumerate(self.rois):
        A[roi, num] = 1
    rval, significant_samples = classify_components_ep(
        self.dataset_trials_filtered_loaded, A,
        np.vstack([self.get_time_trace(x + 1) for x in range(len(self.rois))]))
    pass


def export(self, matlab, background_images, color_maps):
    # temp_type = self.time_trace_params["time_trace_type"]
    # for time_type in ["Mean", "DeltaF Over F"]:
    #     for denoise in [True, False]:
    #         self.time_trace_params["denoise"] = denoise
    #         self.time_trace_params["time_trace_type"] = time_type
    #         self.calculate_time_traces()
    #         # if time_type == "Mean" and denoise:
    #         #     self.calculate_statistics()
    # self.time_trace_params["time_trace_type"] = temp_type
    self.save_rois(self.rois)
    roi_save_object = []
    spatial_box = SpatialBox(0, 1, image_shape=self.shape, spatial_overlap=0)
    rois_all = []
    for class_id in self.classes:
        class_name = self.classes[class_id]["name"]
        for roi in self.classes[class_id]["rois"]:
            rois_all.append((self.rois_dict[roi]["pixels"], class_name))
    for num, (roi, class_name) in enumerate(rois_all):
        if self.dataset_params["crop_stack"]:

            cords = [[x[0] + self.dataset_params["crop_x"][0],
                      x[1] + self.dataset_params["crop_y"][0]] for x in
                     spatial_box.convert_1d_to_2d(roi)]
        else:
            cords = spatial_box.convert_1d_to_2d(roi)
        curr_roi = {"id": num, "coordinates": cords, "class": class_name}
        roi_save_object.append(curr_roi)

    with open(os.path.join(self.save_dir_path, "roi_list.json"), "w") as f:
        json.dump(roi_save_object, f)
    if matlab:
        test = {x[:31].replace(" ", "_"): np.vstack(self.time_traces[x]) for x in
                self.time_traces.keys()}

        savemat(os.path.join(self.save_dir_path, "time_traces.mat"), {"data": test},
                appendmat=True)
        savemat(os.path.join(self.save_dir_path, "rois.mat"), {"data": self.rois},
                appendmat=True)

    shape = self.shape

    if not os.path.isdir(os.path.join(self.save_dir_path, "images/")):
        os.mkdir(os.path.join(self.save_dir_path, "images/"))
    rois = []
    for num, roi in enumerate(self.rois):
        cords = spatial_box.convert_1d_to_2d(roi)
        rois.append(cords)
    fig = plt.figure(frameon=False)
    fig.set_size_inches(shape[1] / 100, shape[0] / 100)

    for background_image_name in background_images:
        try:
            if background_image_name == "mean":
                background_image = np.mean(
                    [self.mean_images[x] for x in self._trials_loaded_indices], axis=0)
            elif background_image_name == "max":
                background_image = np.max(
                    [self.max_images[x] for x in self._trials_loaded_indices], axis=0)
            elif background_image_name == "blank":
                background_image = np.zeros(shape)
            else:  # eigen norm is default
                background_image = create_image_from_eigen_vectors(
                    os.path.join(self.save_dir_path, "eigen_vectors/"), shape)
            if not background_image_name == "zeros":
                background_image = scale_background(background_image)
            np.save(os.path.join(self.save_dir_path, "images/",
                                 "{}.npy".format(background_image_name)), background_image)
            for color_map_name in color_maps:
                # first the blob plot
                ax = plt.Axes(fig, [0., 0., 1., 1.])
                ax.set_axis_off()
                fig.add_axes(ax)
                if color_map_name == "green":
                    ax.imshow(np.dstack(
                        [np.zeros(shape), background_image, np.zeros(shape)]).astype(
                        np.uint8), vmin=0, vmax=255)

                else:
                    ax.imshow(background_image.astype(np.uint8), vmin=0, vmax=255,
                              cmap=color_map_name)
                ax.imshow(create_roi_image_blob(color_list=self.color_list, rois=rois,
                                                shape=shape).astype(np.uint8), vmin=0,
                          vmax=255)
                fig.savefig(os.path.join(self.save_dir_path, "images/",
                                         "roi_blob_%s_image_%s.png" % (
                                         background_image_name, color_map_name)))
                fig.clf()
                # now the outline plot
                ax = plt.Axes(fig, [0., 0., 1., 1.])
                ax.set_axis_off()
                fig.add_axes(ax)
                if color_map_name == "green":
                    ax.imshow(np.dstack(
                        [np.zeros(shape), background_image, np.zeros(shape)]).astype(
                        np.uint8), vmin=0, vmax=255)

                else:
                    ax.imshow(background_image.astype(np.uint8), vmin=0, vmax=255,
                              cmap=color_map_name)
                plot_roi_image_contour(shape, (1, 1, 1), rois, ax)
                fig.savefig(os.path.join(self.save_dir_path, "images/",
                                         "roi_outline_%s_image_%s.png" % (
                                         background_image_name, color_map_name)))
                fig.clf()
                # now just background plot
                ax = plt.Axes(fig, [0., 0., 1., 1.])
                ax.set_axis_off()
                fig.add_axes(ax)
                if color_map_name == "green":
                    ax.imshow(np.dstack(
                        [np.zeros(shape), background_image, np.zeros(shape)]).astype(
                        np.uint8), vmin=0, vmax=255)

                else:
                    ax.imshow(background_image.astype(np.uint8), vmin=0, vmax=255,
                              cmap=color_map_name)
                fig.savefig(os.path.join(self.save_dir_path, "images/",
                                         "background_%s_image_%s.png" % (
                                         background_image_name, color_map_name)))
                fig.clf()
        except ValueError:
            print("Failed to export images please run roi extraction part again if you want to export images")


def save_image(self, image, path):
    plt.imsave(path, image.reshape((self.shape[0], self.shape[1], 3)).astype(np.uint8),
               vmin=0, vmax=255)


def create_image_from_eigen_vectors(path, shape, vectors=None):
    if vectors is None:
        onlyfiles = [f for f in listdir(path) if isfile(join(path, f))]
        vectors = []
        for x in onlyfiles:
            with open(os.path.join(path, x), "rb") as file:
                vectors.append(pickle.load(file)[:, 1:])
        all_vectors = np.hstack(vectors)
    else:
        all_vectors = vectors
    all_vectors_sum = np.power(np.sum(np.power(all_vectors, 2), axis=1), .5)
    all_vectors_shaped = np.reshape(all_vectors_sum, shape)
    all_vectors_shaped[all_vectors_shaped < 0] = 0
    # if all_vectors_shaped.min()<0:
    #     all_vectors_shaped+=all_vectors_shaped.min()*-1
    return all_vectors_shaped * 255 / (all_vectors_shaped.max())


def scale_background(background_image):
    background_image[background_image < 0] = 0

    background_image = (((background_image - np.percentile(background_image, 1)) / (
            np.percentile(background_image, 97) - np.percentile(
        background_image, 1))))

    background_image[background_image > 1] = 1
    background_image = background_image * 255
    background_image[background_image < 0] = 0
    return background_image


def create_roi_image_blob(color_list, rois, shape):
    combined_image = np.dstack([np.zeros(shape), np.zeros(shape),
                                np.zeros(shape)])
    transparency = np.zeros(shape)

    for num, cords in enumerate(rois):
        for pixel in cords:
            combined_image[pixel[0], pixel[1]] = color_list[num % len(color_list)]
            transparency[pixel[0], pixel[1]] = 255

    combined_image = np.dstack([combined_image, transparency])
    return combined_image


def plot_roi_image_contour(shape, color, rois, fig):
    for num, cords in enumerate(rois):

        image_temp = np.zeros((shape[0], shape[1]), dtype=float)

        for pixel in cords:
            image_temp[pixel[0], pixel[1]] = 1

        # edge = feature.canny(
        #     np.sum(image_temp, axis=2) / np.max(np.sum(image_temp, axis=2)))
        # image[edge] = 1
        # image_temp = ndimage.morphology.binary_dilation(image_temp)
        test = measure.label(image_temp, background=0, connectivity=1)
        # image_temp = ndimage.morphology.binary_erosion(image_temp)
        #
        # image_temp = ndimage.morphology.binary_erosion(image_temp)
        # image_temp = ndimage.binary_closing(image_temp)
        # print(test.max())
        for x in range(test.max()):
            image = np.zeros((shape[0], shape[1]), dtype=float)
            image[test == x + 1] = 1
            contour = find_contours(image, .3)
            if len(contour) != 0:
                fig.plot(contour[0][:, 1], contour[0][:, 0], color=color,
                         linewidth=2)
            # plt.imshow(image)
            # fig.plot(contour[0][:, 1], contour[0][:, 0], color=color, linewidth=2)


def export_overlapping_rois(self, current_roi_num_1, current_roi_num_2):
    """THis is only used in dev version it allows the user to export time trace of all pixels in this roi and in overlapping rois"""
    current_roi_num_1 = current_roi_num_1 - 1
    current_roi_num_2 = current_roi_num_2 - 1
    pixels_in_1_not_2 = [x for x in self.rois[current_roi_num_1] if
                         x not in self.rois[current_roi_num_2]]
    pixels_in_2_not_1 = [x for x in self.rois[current_roi_num_2] if
                         x not in self.rois[current_roi_num_1]]
    pixels_in_1_and_2 = [x for x in self.rois[current_roi_num_2] if
                         x in self.rois[current_roi_num_1]]
    time_traces_1_not_2 = []
    time_traces_2_not_1 = []
    time_traces_1_and_2 = []

    for trial_num in self.trials_loaded_time_trace_indices:
        data = self.load_trial_dataset_step(trial_num).compute()
        data_2d = reshape_to_2d_over_time(data[:])
        del data
        time_traces_1_and_2.append(data_2d[pixels_in_1_and_2])
        time_traces_1_not_2.append(data_2d[pixels_in_1_not_2])
        time_traces_2_not_1.append(data_2d[pixels_in_2_not_1])
    time_traces_2_not_1 = np.hstack(time_traces_2_not_1)
    time_traces_1_not_2 = np.hstack(time_traces_1_not_2)
    time_traces_1_and_2 = np.hstack(time_traces_1_and_2)
    out = {"time_trace_2_not_1": time_traces_2_not_1,
           "time_trace_1_not_2": time_traces_1_not_2,
           "time_traces_1_and_2": time_traces_1_and_2}
    pickle_save(out, "roi_overlap.pickle", output_directory=self.save_dir_path)



import logging

from qtpy.QtWidgets import *

from cidan.GUI.Data_Interaction.ROIExtractionThread import ROIExtractionThread
from cidan.GUI.ImageView.ROIImageViewModule import ROIImageViewModule
from cidan.GUI.Inputs.OptionInput import OptionInput
from cidan.GUI.ListWidgets.GraphDisplayWidget import GraphDisplayWidget
from cidan.GUI.ListWidgets.ROIListModule import ROIListModule
from cidan.GUI.ListWidgets.TrialListWidget import TrialListWidget
from cidan.GUI.Tabs.Tab import Tab

logger1 = logging.getLogger("cidan.AnalysisTab")


class AnalysisTab(Tab):
    def __init__(self, main_widget):
        self.main_widget = main_widget
        self.cur_plot_type = "neuron"

        self.image_view = ROIImageViewModule(self.main_widget, self,select_multiple=True, settings_tab=False)

        self.roi_list_module = ROIListModule(main_widget.data_handler, self,
                                             select_multiple=True, display_time=False)
        self.thread = ROIExtractionThread(main_widget, QPushButton(),
                                          self.roi_list_module, self)
        self.main_widget.thread_list.append(self.thread)

        self.update_time = True
        settings_tabs = QTabWidget()
        plot_settings_widget = QWidget()
        plot_settings_layout = QVBoxLayout()
        plot_settings_widget.setLayout(plot_settings_layout)
        settings_tabs.addTab(plot_settings_widget, "Settings")

        self.plot_type_input = OptionInput(display_name="Plot Type", program_name="",
                                           on_change_function=lambda x,
                                                                     y: self.deselectRoiTime(),
                                           default_index=0, tool_tip="",
                                           display_tool_tip=False,
                                           val_list=["Line", "Color Mesh"])
        plot_settings_layout.addWidget(self.plot_type_input)
        self.plot_by_input = OptionInput(display_name="Plot By", program_name="",
                                         on_change_function=lambda x,
                                                                   y: self.deselectRoiTime(),
                                         default_index=0, tool_tip="",
                                         display_tool_tip=False,
                                         val_list=["Neuron", "Trial"])
        plot_settings_layout.addWidget(self.plot_by_input)
        self.time_trace_type = OptionInput("Time Trace Type", "",
                                           lambda x, y: self.deselectRoiTime(),
                                           default_index=0,
                                           tool_tip="Select way to calculate time trace",
                                           val_list=list(
                                               self.data_handler.time_trace_possibilities_functions.keys()))
        plot_settings_layout.addWidget(self.time_trace_type)

        time_trace_settings = QWidget()
        time_trace_settings_layout = QVBoxLayout()
        if not len(self.data_handler.trials_loaded) == 1 and not \
        self.data_handler.dataset_params["single_file_mode"] and not \
        self.data_handler.dataset_params["trial_split"]:
            settings_tabs.addTab(time_trace_settings, "Trial Selector")
        else:
            plot_settings_layout.addWidget(time_trace_settings)
            time_trace_settings.hide()

        settings_tabs.addTab(self.image_view.createSettings(), "Display Settings")
        time_trace_settings_layout.setContentsMargins(0, 0, 0, 0)
        time_trace_settings.setLayout(time_trace_settings_layout)
        self._time_trace_trial_select_list = TrialListWidget(False)
        self._time_trace_trial_select_list.setMinimumHeight(115)
        self._time_trace_trial_select_list.set_items_from_list(
            self.data_handler.trials_all,
            self.data_handler.trials_loaded_time_trace_indices)
        time_trace_settings_layout.addWidget(self._time_trace_trial_select_list,
                                             stretch=5)
        time_trace_update_button = QPushButton("Update Time Traces")
        time_trace_settings_layout.addWidget(time_trace_update_button)
        time_trace_update_button.clicked.connect(
            lambda x: self.update_time_traces())


        self.plot_widget = GraphDisplayWidget(self.main_widget)
        self.deselect_all_button = QPushButton("Deselect All")
        self.deselect_all_button.clicked.connect(lambda x: self.selectAll(False))
        self.select_all_button = QPushButton("Select All")
        self.select_all_button.clicked.connect(lambda x: self.selectAll(True))
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.select_all_button)
        button_layout.addWidget(self.deselect_all_button)
        export_overlapping_rois = QPushButton(text="Export Overlapping ROIs")
        export_overlapping_rois.hide()
        export_overlapping_rois.clicked.connect(lambda: self.export_overlap_rois())
        if self.main_widget.dev:
            export_overlapping_rois.show()


        # bottom_half.addWidget(self.plot_widget, stretch=2)
        if self.main_widget.data_handler.rois_loaded:
            self.updateTab()
            # self.plot_widget.set_list_items([self.data_handler.get_time_trace(x) for x in range(20)], [x for x in range(20)], None)
        super().__init__("Analysis",
                         column_1=[self.roi_list_module, button_layout,
                                   export_overlapping_rois, settings_tabs],
                         column_2=[self.image_view, self.plot_widget],
                         column_2_display=True, column2_moveable=True)

    def selectAll(self, select):
        self.update_time = False
        for x in self.roi_list_module.roi_item_list:
            x.check_box.setChecked(select)
        self.update_time = True
        self.deselectRoiTime()
    def selectRoiTime(self, num):
        self.deselectRoiTime()

    def deselectRoiTime(self):
        if any(self.data_handler.roi_time_trace_need_update):
            self.main_widget.console.updateText(
                "Some time traces are out of date, please recalculate",
                warning=True)
        if self.update_time:
            if self.plot_by_input.current_state() == "Neuron":
                if self.cur_plot_type != "neuron":
                    self.cur_plot_type = "neuron"
                    self.roi_list_module.select_multiple = True

                try:
                    data_list = []
                    roi_names = []
                    for num2, x in zip(
                            range(0, len(
                                self.roi_list_module.roi_time_check_list)),
                            self.roi_list_module.roi_time_check_list):
                        if x:
                            data_list.append(
                                self.main_widget.data_handler.get_time_trace(num2,
                                                                             trace_type=self.time_trace_type.current_state()))
                            roi_names.append(
                                self.data_handler.roi_index_backward[num2])
                    self.plot_widget.set_list_items(data_list, roi_names, [],
                                                    p_color=self.plot_type_input.current_state() == "Color Mesh",
                                                    type="neuron")
                except AttributeError as e:
                    print("No ROIs have been generated yet")
            else:
                if self.cur_plot_type != "trial":
                    self.roi_list_module.select_multiple = False
                    self.cur_plot_type = "trial"
                    self.selectAll(False)

                try:
                    print(self.roi_list_module.current_selected_roi)
                    if self.roi_list_module.current_selected_roi is not None:
                        roi = self.roi_list_module.current_selected_roi
                        data_list = []
                        roi_names = [self.data_handler.rois_index_backward[roi]]
                        for x in self.data_handler.trials_loaded_time_trace_indices:
                            data_list.append(
                                self.main_widget.data_handler.get_time_trace(roi,
                                                                             x))
                        self.plot_widget.set_list_items(data_list, roi_names,
                                                        self.data_handler.trials_loaded_time_trace_indices,
                                                        p_color=self.plot_type_input.current_state() == "Color Mesh",
                                                        type="trial")
                    else:
                        self.plot_widget.set_list_items([], [],
                                                        [],
                                                        p_color=self.plot_type_input.current_state() == "Color Mesh",
                                                        type="trial")
                except AttributeError:
                    print("No ROIs have been generated yet")

    @property
    def data_handler(self):
        return self.main_widget.data_handler

    def update_time_traces(self):
        if (self.main_widget.checkThreadRunning()):
            self.data_handler.calculate_time_traces()
            self.data_handler.update_selected_trials(
                self._time_trace_trial_select_list.selectedTrials())
            self.deselectRoiTime()

    def updateTab(self):
        if (self.main_widget.checkThreadRunning()):
            self._time_trace_trial_select_list.set_items_from_list(
                self.data_handler.trials_all,
                self.data_handler.trials_loaded_time_trace_indices)
            if self.data_handler.rois_loaded:
                self.roi_list_module.set_list_items(
                    self.main_widget.data_handler.rois_dict)
                self.selectAll(False)
                self.image_view.reset_view()
                try:
                    # self.update_time = False
                    for x in self.roi_list_module.roi_item_list[:1]:
                        x.check_box.setChecked(True)
                    # self.deselectRoiTime()
                    # self.update_time = True
                except:
                    pass
            else:
                self.image_view.reset_view()

    def export_overlap_rois(self):
        roi_nums = []
        for num, x in zip(range(1, len(self.roi_list_module.roi_time_check_list) + 1),
                          self.roi_list_module.roi_time_check_list):
            if x:
                roi_nums.append(num)
        if len(roi_nums) == 2:
            self.data_handler.export_overlapping_rois(roi_nums[0], roi_nums[1])
            print("export success")



from qtpy import QtCore
from qtpy.QtWidgets import *


class ClassItemWidget(QWidget):
    """
    Item in the Class list, takes care of everything except for color part which is
    handled by ClassItemModule
    """

    def __init__(self, classifier_tab, color, class_dict, class_num, name, parent=None,
                 id=id):
        self.classifier_tab = classifier_tab
        self.class_dict = class_dict
        self.name = name
        self.id = id
        self.class_num = class_num
        super(ClassItemWidget, self).__init__(parent)
        self.setStyleSheet("""QPushButton {background-color: rgba(0,0,0,0%);
        padding-left:3px;
        padding-right:3px;
        
        color: #CCCCCC;}
        QPushButton:hover {
          border: 1px solid #148CD2;
          background-color: #505F69;
          color: #F0F0F0;
            }
            QPushButton:pressed {
              background-color: #19232D;
              border: 1px solid #19232D;
            }
            
            QPushButton:pressed:hover {
              border: 1px solid #148CD2;
            }
            QPushButton:selected {
              background-color: rgba(0,0,0,0%);
              color: #32414B;
            }
            QLabel {
            background-color: rgba(0,0,0,0%)
            }QCheckBox {
            background-color: rgba(0,0,0,0%)
            }""")
        self.edit_class = QPushButton("Edit Class")
        self.edit_class.clicked.connect(
            lambda x: self.classifier_tab.edit_class(self.id))
        # self.check_box = QCheckBox()
        # self.check_box.toggled.connect(lambda: self.check_box_toggled())
        # self.check_box_time_trace = QCheckBox()
        # self.check_box_time_trace.toggled.connect(lambda: self.time_check_box_toggled())

        lay = QHBoxLayout(self)
        lay.addWidget(QLabel(self.name), alignment=QtCore.Qt.AlignLeft)
        lay.addWidget(QLabel(text="#" + str(class_num)), alignment=QtCore.Qt.AlignLeft)
        # if display_time:
        lay.addWidget(QLabel())
        lay.addWidget(QLabel())
        lay.addWidget(QLabel())
        # lay.addWidget(
        #     QLabel(str(round(self.classifier_tab.data_handler.class_circ_list[class_num - 1], 3))))
        lay.addWidget(self.edit_class)
        lay.setContentsMargins(0, 0, 0, 0)

    def keyPressEvent(self, event):
        self.classifier_tab.keyPressEvent(event)
    # def select_check_box(self):
    #     if not self.check_box.checkState():
    #         if not self.class_dict.select_multiple:
    #             for x in self.class_dict.class_item_list:
    #                 if x != self:
    #                     x.check_box.setChecked(False)
    #         self.check_box.setChecked(True)
    #         if not self.display_time:
    #             self.check_box_time_trace.setChecked(True)
    #         self.class_dict.current_selected_class = self.class_num
    #
    #     else:
    #         self.check_box.setChecked(False)
    #         if not self.display_time:
    #             self.check_box_time_trace.setChecked(False)
    #         self.class_dict.current_selected_class = None
    #
    # def selected(self):
    #     return self.check_box.checkState()
    # def select_time_check_box(self):
    #     self.check_box_time_trace.setChecked(not self.check_box_time_trace.checkState())
    #
    # def check_box_toggled(self):
    #     if self.check_box.checkState():
    #
    #         if not self.class_dict.select_multiple:
    #             for x in self.class_dict.class_item_list:
    #                 if x != self:
    #                     x.check_box.setChecked(False)
    #
    #         self.class_dict.current_selected_class = self.class_num
    #         self.check_box_time_trace.setChecked(True)
    #
    #         if not self.display_time:
    #             self.classifier_tab.image_view.selectRoi(self.class_num)
    #     else:
    #         self.class_dict.current_selected_class = None
    #         if not self.display_time:
    #             self.check_box_time_trace.setChecked(False)
    #         self.classifier_tab.image_view.deselectRoi(self.class_num)
    #
    # def time_check_box_toggled(self):
    #
    #     self.class_dict.class_time_check_list[
    #         self.class_num - 1] = self.check_box_time_trace.checkState()
    #     if self.check_box_time_trace.checkState():
    #         self.classifier_tab.selectRoiTime(self.class_num)
    #     else:
    #         self.classifier_tab.deselectRoiTime()



import logging

import dask
import numpy as np

from cidan.LSSC.functions.data_manipulation import reshape_to_2d_over_time
from cidan.LSSC.functions.pickle_funcs import *
from cidan.LSSC.functions.progress_bar import printProgressBar
from cidan.TimeTrace.neuropil import calculate_neuropil
from cidan.TimeTrace.waveletDenoise import waveletDenoise

logger1 = logging.getLogger("cidan.DataHandler")


def calculate_time_traces(self, report_progress=None):
    """
    Calculates the time traces for every roi in self.rois
    """
    self.neuropil_pixels = calculate_neuropil(image_shape=self.shape,
                                              roi_list=self.rois,
                                              roi_mask_flat=self.pixel_with_rois_flat,
                                              min_pixels=100)  # self.time_trace_params["min_neuropil_pixels"])
    self.neuropil_image_display = np.zeros(
        [self.shape[0] * self.shape[1]])
    for neurolpil in self.neuropil_pixels:
        self.neuropil_image_display[neurolpil] = 255
    self.neuropil_image_display = np.reshape(
        np.dstack([self.neuropil_image_display, np.zeros(
            [self.shape[0] * self.shape[1]]),
                   (self.pixel_with_rois_flat > 0) * 255]),
        [self.shape[0] * self.shape[1], 3])
    self.roi_time_trace_need_update = []
    for _ in range(len(self.rois)):
        self.roi_time_trace_need_update.append(False)
    self.time_traces = {x: [] for x in
                        list(self.time_trace_possibilities_functions)}
    for x in self.time_traces.keys():
        for _ in range(len(self.rois)):
            self.time_traces[x].append([])
            for _ in range(1 if self.single_dataset_mode else len(self.trials_all)):
                self.time_traces[x][-1].append(False)
    calc_list = []
    roi_time_traces_by_pixel = []
    for _ in range(len(self.rois)):
        roi_time_traces_by_pixel.append([])
        for _ in range(1 if self.single_dataset_mode else len(self.trials_all)):
            roi_time_traces_by_pixel[-1].append(False)
    roi_neuropil_traces_by_pixel = []
    for _ in range(len(self.rois)):
        roi_neuropil_traces_by_pixel.append([])
        for _ in range(1 if self.single_dataset_mode else len(self.trials_all)):
            roi_neuropil_traces_by_pixel[-1].append(False)


    if not self.real_trials:
        calc_list = []
        data_2d = reshape_to_2d_over_time(self.dataset_list[0][:])
        sub_2_15 = np.isclose(np.mean(data_2d[:50,:5]), 2 ** 15, 0, 4000)
        def preprocess(data):
            if sub_2_15:
                data = data - 2 ** 15
                data[data < -400] = np.min(data[data > -400])
            return data
        for roi in range(len(self.rois)):
            roi_time_traces_by_pixel[roi][0] = preprocess(data_2d[self.rois[roi]].astype(np.float32))
            roi_neuropil_traces_by_pixel[roi][0] =preprocess(data_2d[self.neuropil_pixels[roi]].astype(np.float32))
        if report_progress is not None:
            printProgressBar(self.trials_loaded_time_trace_indices[-1],
                             total=len(
                                 self.trials_loaded_time_trace_indices) + len(
                                 self.rois) + 2,
                             prefix="Time Trace Calculation Progress:",
                             suffix="Complete", progress_signal=report_progress)
        roi_time_traces_by_pixel_denoised = dask.compute(*[dask.delayed(waveletDenoise)(x[0]) for x in roi_time_traces_by_pixel])
        roi_neuropil_traces_by_pixel_denoised = dask.compute(*[dask.delayed(waveletDenoise)(x[0]) for x in roi_neuropil_traces_by_pixel])

        for roi_counter, roi_data, neuropil_data,roi_data_denoised_combined,neuropil_data_denoised_combined in zip(range(len(self.rois)),
                                                        roi_time_traces_by_pixel,
                                                        roi_neuropil_traces_by_pixel,roi_time_traces_by_pixel_denoised,roi_neuropil_traces_by_pixel_denoised):
            roi_data_combined = roi_data[0]
            neuropil_data_combined = neuropil_data[0] #np.vstack([x[0] for x in roi_time_traces_by_pixel])

            for key in self.time_traces.keys():
                if "Denoised" in key:
                    self.time_traces[key][roi_counter] = [
                        self.time_trace_possibilities_functions[key](
                            roi_data_denoised_combined,
                            neuropil_data_denoised_combined,
                        )]
                else:
                    self.time_traces[key][roi_counter] = [
                        self.time_trace_possibilities_functions[key](
                            roi_data_combined, neuropil_data_combined)]
            if report_progress is not None:
                printProgressBar(
                    len(
                        self.trials_loaded_time_trace_indices) + roi_counter + 1,
                    total=len(
                        self.trials_loaded_time_trace_indices) + len(self.rois) + 2,
                    prefix="Time Trace Calculation Progress:",
                    suffix="Complete", progress_signal=report_progress)
    if self.real_trials:
        for trial_num in self.trials_loaded_time_trace_indices:
            data =self.dataset_list[trial_num]
            data_2d = reshape_to_2d_over_time(data[:]).astype(np.int32)-2**15
            data_2d[data_2d<-400]=np.min(data_2d[data_2d>=-400])
            del data
            # if type(self.dataset_trials_filtered[trial_num]) == bool:
            #     data = self.load_trial_filter_step(
            #         trial_num).compute()
            #     self.dataset_trials_filtered[trial_num] = data
            #
            #     del data
            # else:
            #     data_2d = reshape_to_2d_over_time(
            #         self.dataset_trials_filtered[trial_num])[:]
            calc_list = []
            for roi in range(len(self.rois)):
                roi_time_traces_by_pixel[roi][trial_num] = data_2d[self.rois[roi]]
                roi_neuropil_traces_by_pixel[roi][trial_num] = data_2d[
                    self.neuropil_pixels[roi]]
            if report_progress is not None:
                printProgressBar(self.trials_loaded_time_trace_indices.index(trial_num),
                                 total=len(
                                     self.trials_loaded_time_trace_indices) + len(
                                     self.rois) + 2,
                                 prefix="Time Trace Calculation Progress:",
                                 suffix="Complete", progress_signal=report_progress)
        for roi_counter, roi_data, neuropil_data in zip(range(len(self.rois)),
                                                        roi_time_traces_by_pixel,
                                                        roi_neuropil_traces_by_pixel):
            roi_data_denoised = [waveletDenoise(x) if type(x) != bool else False for
                                 x in roi_data]
            neuropil_data_denoised = [
                waveletDenoise(x) if type(x) != bool else False for
                x in neuropil_data]
            for key in self.time_traces.keys():
                if "Denoised" in key:
                    for trial_num in self.trials_loaded_time_trace_indices:
                        self.time_traces[key][roi_counter][trial_num] = \
                            self.time_trace_possibilities_functions[key](
                                roi_data_denoised[trial_num],
                                neuropil_data_denoised[trial_num])
                else:
                    for trial_num in self.trials_loaded_time_trace_indices:
                        self.time_traces[key][roi_counter][trial_num] = \
                            self.time_trace_possibilities_functions[key](
                                roi_data[trial_num], neuropil_data[trial_num])
            if report_progress is not None:
                printProgressBar(
                    len(
                        self.trials_loaded_time_trace_indices) + roi_counter + 1,
                    total=len(
                        self.trials_loaded_time_trace_indices) + len(self.rois) + 2,
                    prefix="Time Trace Calculation Progress:",
                    suffix="Complete", progress_signal=report_progress)
    self.roi_time_trace_need_update = [False for _ in
                                       self.roi_time_trace_need_update]

    if os.path.isdir(self.save_dir_path):
        pickle_save(self.time_traces,
                    "time_traces",
                    output_directory=self.save_dir_path)
    self.rois_loaded = True


# def calculate_time_trace(self, roi_num, trial_num=None, data_2d=None):
#     """
#     Calculates a time trace for a certain ROI and save to time trace list
#     Parameters
#     ----------
#     roi_num
#         roi to calculate for this starts at [1..number of rois]
#     trial_num
#         indice of trial in trials_all starts at [0, number of trials-1] if
#         none then calculates for all trials
#     """
#     trial_nums = [trial_num]
#     if trial_num == None:
#         trial_nums = self.trials_loaded_time_trace_indices
#     for trial_num in trial_nums:
#         roi = self.rois[roi_num - 1]
#
#         if type(self.dataset_trials_filtered[
#                     trial_num]) == bool and data_2d == None:
#             self.dataset_trials_filtered[trial_num] = self.load_trial_filter_step(
#                 trial_num).compute()
#         if data_2d is None:
#             data_2d = reshape_to_2d_over_time(
#                 self.dataset_trials_filtered[trial_num])
#         if self.time_trace_params[
#             "time_trace_type"] == "DeltaF Over F" and self.real_trials:
#             time_trace = calculateDeltaFOverF(roi, data_2d,
#                                               denoise=self.time_trace_params[
#                                                   "denoise"] if self.real_trials else False)
#         else:
#             time_trace = calculateMeanTrace(roi, data_2d,
#                                             denoise=self.time_trace_params[
#                                                 "denoise"] if self.real_trials else False)
#         self.time_traces[roi_num - 1][trial_num] = time_trace

def get_time_trace(self, num, trial=None, trace_type="Mean Florescence"):
    """
    Returns the time trace for a certain roi over all currently selected trials
    Parameters
    ----------
    num : int
        ROI num starts at 0 to num rois

    Returns
    -------
    np.ndarray of the time trace
    """
    if not self.real_trials:
        try:
            if len(self.time_traces[trace_type][num]) == 1 and type(
                    self.time_traces[trace_type][num][0]) != bool:
                return self.time_traces[trace_type][num][0]
            else:
                return False

        except IndexError:
            return False
    try:
        if (trial == None):

            if self.real_trials:
                output = np.ndarray(shape=(0))
                for trial_num in self.trials_loaded_time_trace_indices:
                    output = np.hstack(
                        [output, self.time_traces[trace_type][num][trial_num]])
            else:
                return self.time_traces[num]
        else:
            output = self.time_traces[num][trial]
    except IndexError:
        return False
    return output



from qtpy import QtCore
from qtpy.QtWidgets import *


class ROIItemWidget(QWidget):
    """
    Item in the ROI list, takes care of everything except for color part which is
    handled by ROIItemModule
    """

    def __init__(self, roi_tab, color, roi_list, id, roi_num, parent=None,
                 display_time=True):
        self.roi_tab = roi_tab
        self.roi_list = roi_list
        self.display_time = display_time
        self.roi_num = roi_num
        self.id = id
        super(ROIItemWidget, self).__init__(parent)
        self.setStyleSheet("""QPushButton {background-color: rgba(0,0,0,0%);
        padding-left:3px;
        padding-right:3px;
        
        color: #CCCCCC;}
        QPushButton:hover {
          border: 1px solid #148CD2;
          background-color: #505F69;
          color: #F0F0F0;
            }
            QPushButton:pressed {
              background-color: #19232D;
              border: 1px solid #19232D;
            }
            
            QPushButton:pressed:hover {
              border: 1px solid #148CD2;
            }
            QPushButton:selected {
              background-color: rgba(0,0,0,0%);
              color: #32414B;
            }
            QLabel {
            background-color: rgba(0,0,0,0%)
            }QCheckBox {
            background-color: rgba(0,0,0,0%)
            }""")
        self.zoom_button = QPushButton("Zoom To")
        self.zoom_button.clicked.connect(
            lambda x: self.roi_tab.image_view.zoomRoi(self.id, input_key=True))
        self.check_box = QCheckBox()
        self.check_box.toggled.connect(lambda: self.check_box_toggled())
        self.check_box_time_trace = QCheckBox()
        self.check_box_time_trace.toggled.connect(lambda: self.time_check_box_toggled())

        lay = QHBoxLayout(self)
        lay.addWidget(self.check_box, alignment=QtCore.Qt.AlignLeft)
        lay.addWidget(QLabel(text="#" + str(id)), alignment=QtCore.Qt.AlignLeft)
        if display_time:
            lay.addWidget(QLabel())
            lay.addWidget(QLabel())
            lay.addWidget(QLabel())
        # lay.addWidget(
        #     QLabel(str(round(self.roi_tab.data_handler.roi_circ_list[roi_num - 1], 3))))
        lay.addWidget(self.zoom_button)
        if display_time:
            lay.addWidget(self.check_box_time_trace, alignment=QtCore.Qt.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

    def keyPressEvent(self, event):
        self.roi_tab.keyPressEvent(event)

    def select_check_box(self, force_on=False):
        if not self.check_box.checkState() or force_on:
            if not self.roi_list.select_multiple:
                for x in self.roi_list.roi_item_list:
                    if x != self:
                        x.check_box.setChecked(False)

            self.check_box.setChecked(True)

            if not self.display_time:
                self.check_box_time_trace.setChecked(True)
            self.roi_list.current_selected_roi = self.roi_num
            try:
                self.roi_tab.update_current_roi_selected()
            except AttributeError:
                pass

        else:
            self.check_box.setChecked(False)
            if not self.display_time:
                self.check_box_time_trace.setChecked(False)
            self.roi_list.current_selected_roi = None
            try:
                self.roi_tab.update_current_roi_selected()
            except AttributeError:
                pass
        self.roi_list.update_select_number()

    def selected(self):
        return self.check_box.checkState()
    def select_time_check_box(self):
        self.check_box_time_trace.setChecked(not self.check_box_time_trace.checkState())

    def check_box_toggled(self):
        if self.check_box.checkState():

            if not self.roi_list.select_multiple:
                for x in self.roi_list.roi_item_list:
                    if x != self:
                        x.check_box.setChecked(False)

            self.roi_list.current_selected_roi = self.roi_num
            try:
                self.roi_tab.update_current_roi_selected()
            except AttributeError:
                pass
            self.check_box_time_trace.setChecked(True)

            if not self.display_time:
                self.roi_tab.image_view.selectRoi(self.roi_num)
        else:
            self.roi_list.current_selected_roi = None
            try:
                self.roi_tab.update_current_roi_selected()
            except AttributeError:
                pass
            if not self.display_time:
                self.check_box_time_trace.setChecked(False)
            self.roi_tab.image_view.deselectRoi(self.roi_num, other_selected=self.roi_list.currently_selected_rois_list)
        self.roi_list.update_select_number()

    def time_check_box_toggled(self):

        self.roi_list.roi_time_check_list[
            self.roi_num] = self.check_box_time_trace.checkState()
        try:
            if self.check_box_time_trace.checkState():
                self.roi_tab.selectRoiTime(self.roi_num)
            else:
                self.roi_tab.deselectRoiTime()
        except AttributeError:
            pass



import json
import os

from numpy import inf, NaN, newaxis, argmin, asarray, isnan, sum, nanmean
from regional import many
from scipy.spatial.distance import cdist


def load(file):
    """
    Load neuronal regions from a file or string.
    """
    if os.path.isfile(file):
        with open(file, 'r') as f:
            values = json.load(f)
    else:
        values = json.loads(file)

    return many([v['coordinates'] for v in values])


def match(a, b, threshold=inf):
    """
    Find unique matches between two sets of regions.

    Params
    ------
    a, b : regions
        The regions to match.

    threshold : scalar, optional, default = inf
        Threshold distance to use when selecting matches.
    """
    targets = b.center
    target_inds = range(0, len(targets))
    matches = []
    for s in a:
        update = 1

        # skip if no targets left, otherwise update
        if len(targets) == 0:
            update = 0
        else:
            dists = cdist(targets, s.center[newaxis])
            if dists.min() < threshold:
                ind = argmin(dists)
            else:
                update = 0

        # apply updates, otherwise add a nan
        if update == 1:
            matches.append(target_inds[ind])
            # targets = delete(targets, ind, axis=0)
            # target_inds = delete(target_inds, ind)
        else:
            matches.append(NaN)

    return matches


def shapes(a, b, threshold=inf):
    """
    Compare shapes between two sets of regions.

    Parameters
    ----------
    a, b : regions
        The regions for which to estimate overlap.

    threshold : scalar, optional, default = inf
        Threshold distance to use when matching indices.
    """

    inds = match(a, b, threshold=threshold)
    d = []
    for jj, ii in enumerate(inds):
        if ii is not NaN:
            d.append(a[jj].overlap(b[ii], method='rates'))
        else:
            d.append((NaN, NaN))

    result = asarray(d)

    if sum(~isnan(result)) > 0:
        inclusion, exclusion = tuple(nanmean(result, axis=0))
    else:
        inclusion, exclusion = 0.0, 0.0

    return inclusion, exclusion


def centers(a, b, threshold=inf):
    """
    Compare centers between two sets of regions.

    The recall rate is the number of matches divided by the number in self,
    and the precision rate is the number of matches divided by the number in other.
    Typically a is ground truth and b is an estimate.
    The F score is defined as 2 * (recall * precision) / (recall + precision)

    Before computing metrics, all sources in self are matched to other,
    and a threshold can be set to control matching.

    Parameters
    ----------
    a, b : regions
        The regions for which to estimate overlap.

    threshold : scalar, optional, default = 5
        The distance below which a source is considered found.
    """
    inds = match(a, b, threshold=threshold)

    d = []
    for jj, ii in enumerate(inds):
        if ii is not NaN:
            d.append(a[jj].distance(b[ii]))
        else:
            d.append(NaN)

    result = asarray(d)

    result[isnan(result)] = inf
    compare = lambda x: x < threshold

    recall = sum(asarray(list(map(compare, result)))) / float(a.count)
    precision = sum(asarray(list(map(compare, result)))) / float(b.count)

    return recall, precision



import shutil
import time

import neurofinder
from PySide2.QtWidgets import QApplication
from skimage import io
from skimage.metrics import structural_similarity

from cidan.GUI.Data_Interaction.loadDataset import load_new_dataset
from cidan.GUI.MainWindow import MainWindow


def test_accuracy():
    app = QApplication([])
    app.setApplicationName("cidan")
    widget = MainWindow(dev=True, preload=False)

    main_widget = widget.table_widget
    load_new_dataset(main_widget, file_path="test_files/small_dataset1.tif",
                     save_dir_path="test_files/save_dir", load_into_mem=True)
    main_widget.open_dataset_thread.wait()
    time.sleep(10)
    data_handler = main_widget.data_handler

    data_handler.change_filter_param("median_filter", True)
    data_handler.change_roi_extraction_param("roi_circ_threshold", 20)
    main_widget.thread_list[2].run()
    # main_widget.thread_list[2].run()
    assert main_widget.tabs[1].image_view.magic_wand(70, 127)

    data_handler.export()
    a = neurofinder.load("test_files/roi_list.json")
    b = neurofinder.load("test_files/save_dir/roi_list.json")
    test = neurofinder.match(a, b)
    percision, recall = neurofinder.centers(a, b)
    inclusion, exclusion = neurofinder.shapes(a, b)
    assert percision > .8
    assert recall > .8
    assert exclusion > .6
    assert inclusion > .6
    image_good = io.imread("test_files/embedding_norm_image.png")
    image_test = io.imread(
        "test_files/save_dir/embedding_norm_images/embedding_norm_image.png")

    assert structural_similarity(image_good, image_test) > .90
    data_handler.change_box_param("total_num_spatial_boxes", 4)
    main_widget.thread_list[1].run()
    data_handler.export()
    image_test = io.imread(
        "test_files/save_dir/embedding_norm_images/embedding_norm_image.png")

    assert structural_similarity(image_good, image_test) > .60
    a = neurofinder.load("test_files/roi_list.json")
    b = neurofinder.load("test_files/save_dir/roi_list.json")
    percision, recall = neurofinder.centers(a, b)
    inclusion, exclusion = neurofinder.shapes(a, b)
    assert percision > .4
    assert recall > .4
    assert exclusion > .4
    assert inclusion > .3
    app.quit()
    shutil.rmtree("test_files/save_dir")



from qtpy.QtCore import QObject, QEvent


class EventFilterMainWindow(QObject):
    def __init__(self, main_window):
        QObject.__init__(self)
        self.main_window = main_window

    def eventFilter(self, obj, event):
        if (event.type() == QEvent.KeyPress):
            key = event.key()
            if self.main_window.checkThreadRunning and self.main_window.data_handler is not None:
                current_active_main_tab_ind = self.main_window.tab_widget.currentIndex()
                current_active_main_tab = self.main_window.tabs[
                    current_active_main_tab_ind - 1]  # since dataload tab isn't in this list
                if hasattr(current_active_main_tab, "keyPressAction"):
                    return current_active_main_tab.keyPressAction(event)
        return False



import time

import numpy as np
from scipy import ndimage
from sklearn.decomposition import NMF

from cidan.LSSC.functions.data_manipulation import reshape_to_2d_over_time, save_image


# TODO Scale A to 0-1
def calculate_spatial_time_denoising(dataset, rois, eigen_norm_image):
    time_all = time.time()

    dataset = dataset.astype(np.float32)
    roi_images = np.zeros(
        [len(rois), eigen_norm_image.shape[0], eigen_norm_image.shape[1]], dtype=int)
    for num, pixels in enumerate(rois):
        for x in pixels:
            roi_images[num, x[0], x[1]] = 1
    A = np.vstack([eigen_norm_image.reshape(
        [1, eigen_norm_image.shape[0], eigen_norm_image.shape[1]]) for _ in rois])
    save_image(A.sum(axis=0), "A.png")

    min_vals = np.amin(A.reshape([A.shape[0], -1]), where=roi_images.reshape([A.shape[0], -1]) == 1, initial=1, axis=1)
    max_vals = np.amax(A.reshape([A.shape[0], -1]), where=roi_images.reshape([A.shape[0], -1]) == 1, initial=0, axis=1)
    A = (A - min_vals.reshape((A.shape[0], 1, 1))) / (
                max_vals.reshape((A.shape[0], 1, 1)) - min_vals.reshape((A.shape[0], 1, 1)))
    A[roi_images == 0] = 0
    # roi_mask_neuropil = np.zeros(
    #     [eigen_norm_image.shape[0] * eigen_norm_image.shape[1]], dtype=int)
    #
    # neuropil_all = np.hstack([x[0] for x in neuropil])
    test = np.sum(roi_images, axis=0)

    test[test > 1] = 1
    roi_mask_neuropil = ndimage.morphology.binary_dilation(test,
                                                           iterations=6)  # image_temp = ndimage.morphology.binary_dilation(image_temp)

    roi_mask_neuropil = roi_mask_neuropil.reshape(
        [eigen_norm_image.shape[0], eigen_norm_image.shape[1]])
    save_image(roi_mask_neuropil, "roi_neuropil.png")

    roi_mask_neuropil = roi_mask_neuropil + np.sum(roi_images, axis=0)

    for x in range(roi_images.shape[0]):
        save_image(roi_images[x], "rois/roi%s.png" % str(x))
    save_image(roi_mask_neuropil, "roi_mask_neuropil.png")
    save_image(np.sum(dataset, axis=0), "dataset.png")
    dataset_masked = np.copy(dataset).transpose([1, 2, 0])

    dataset_masked[roi_mask_neuropil >= 1] = 0
    dataset_masked = dataset_masked.transpose([2, 0, 1])
    save_image(np.sum(dataset_masked, axis=0), "dataset_masked.png")
    # print("data_mask", time.time()-time_all)
    dataset_masked_r = reshape_to_2d_over_time(dataset_masked)
    model = NMF(n_components=1, init='custom', alpha=0, solver='mu',
                beta_loss='kullback-leibler')
    A_mask = time_thing(model.fit_transform)(dataset_masked_r, 1,
                                             np.mean(dataset_masked_r, axis=1).reshape(-1, 1),
                                             np.mean(dataset_masked_r, axis=0).reshape(1, -1))
    C_b = model.components_
    save_image(A_mask.reshape([eigen_norm_image.shape[0], eigen_norm_image.shape[1]]),
               "A_background_mask.png")
    # print("A_background_mask",time.time()-time_all)

    dataset_2d = reshape_to_2d_over_time(dataset)

    pixels_to_calculate = np.nonzero(np.sum(reshape_to_2d_over_time(roi_images[:, :, :]), axis=1))[0]
    C, _ = iterative_solve(dataset_2d, A=A, C_b=C_b,
                           pixels_to_calculate=pixels_to_calculate)
    # total_list = deconstruct_matrix(dataset_2d[pixels_to_calculate,:].transpose(), C_b.transpose())
    # A_mask[pixels_to_calculate] = total_list.transpose()
    #
    # A_b = A_mask
    # A_b = A_b.reshape([eigen_norm_image.shape[0], eigen_norm_image.shape[1]])
    #
    # save_image(A_b, "A_background.png")
    # print("A_background",time.time()-time_all)
    # save_image(np.mean(dataset_2d - A_b.reshape([-1, 1]) * C_b, axis=1).reshape([eigen_norm_image.shape[0], eigen_norm_image.shape[1]]), "A_wo_background_mean.png")
    #
    # C = deconstruct_matrix(dataset_2d - A_b.reshape([-1, 1]) * C_b,reshape_to_2d_over_time(A))
    # dask.compute([dask.delayed(optimize.nnls)(
    # dataset_2d - A_b.reshape([-1, 1]) * C_b, reshape_to_2d_over_time(A)[:, x]) for x
    #               in range(A.shape[0])])
    # C = np.vstack([x[0] for x in C[0]]).transpose()
    # print("All", time.time()-time_all)
    return C, C_b


def iterative_solve(dataset_2d, A, C_b, pixels_to_calculate):
    dataset = dataset_2d[pixels_to_calculate, :]
    A = reshape_to_2d_over_time(A)[pixels_to_calculate, :]

    # A_non_zero = A[A!=0]
    # A = A*4/np.percentile(A_non_zero,2)
    A_non_zero = A[A != 0]

    print("Apercent", np.percentile(A_non_zero, 5), np.percentile(A_non_zero, 15),
          np.percentile(A_non_zero, 25), np.percentile(A_non_zero, 35),
          np.percentile(A_non_zero, 45), np.percentile(A_non_zero, 45))
    C = np.zeros((A.shape[1], dataset.shape[1]))
    for x in range(A.shape[1]):
        selected_A = A[:, x]
        C[x] = np.mean(dataset[selected_A != 0], axis=0).reshape((-1, 1)).transpose()
    C = filter_close_0(C)
    # A_b = filter_close_0(np.random.normal(size=[1,len(pixels_to_calculate)],loc=0.2, scale=.1))
    A_b = filter_close_0(np.mean(dataset - np.matmul(A, C), axis=1).reshape([1, -1])).transpose()

    print(np.mean(C), np.mean(A_b))
    print("A_b", np.count_nonzero(A_b))
    print("C", np.count_nonzero(C))
    A_all = np.hstack([A, A_b])
    C_all = np.vstack([C, C_b])
    for x in range(100):
        numerator_c = np.matmul(A_all.transpose(),
                                dataset)
        denomeinator_part_1_c = np.matmul(A_all.transpose(),
                                          A_all)  # write out in comments what each matmul dimensions should be
        denomeinator_c = np.matmul(denomeinator_part_1_c, C_all)
        C_all = np.divide(np.multiply(C_all, numerator_c), denomeinator_c)
        C_all[-1:] = C_b
        print("C", C.size - np.count_nonzero(C))
        numerator_ab = np.matmul(dataset, C_all.transpose())
        denomeinator_part_1_ab = np.matmul(C_all, C_all.transpose())
        denomeinator_ab = np.matmul(A_all, denomeinator_part_1_ab)
        A_all = np.divide(np.multiply(A_all, numerator_ab), denomeinator_ab)
        A_all[:, 0:A.shape[1]] = A
        # print("A_b", A_b.size-np.count_nonzero(A_b))
    # print(np.mean(C), np.mean(A_b))
    return C_all[0:C.shape[0]], A_all[:, -1:]


def filter_close_0(data):
    data[data < 0.00000000001] = 0.00000000001
    return data


def filter_0(data):
    data[data < 0.00] = 0.00
    return data


def deconstruct_matrix(V, W, dif=1e-15):
    # V[V<0.00000000]=0.00000000
    H = np.zeros((W.shape[1], V.shape[1]))
    for x in range(W.shape[1]):
        selected_W = W[:, x]
        H[x] = np.mean(
            np.divide(V[selected_W != 0], W[:, x][selected_W != 0].reshape((-1, 1))),
            axis=0).reshape((-1, 1)).transpose()
    # cur_w = np.ones_like(W)

    top = np.matmul(W.transpose(), V)
    bottom_part_1 = np.matmul(W.transpose(), W)
    for x in range(5):
        bottom = np.matmul(bottom_part_1, H)
        H_next = np.divide(np.multiply(H, numerator), bottom)
        if np.max(np.abs(H_next - H)) <= dif:
            print(x)
            break
        H = H_next
    print(3000)
    return H_next


def time_thing(func):
    def time_inner(*args):
        time_start = time.time()
        temp = func(*args)
        print(time_start - time.time())
        return temp

    return time_inner



import numpy as np
from pyqtgraph import ImageView, PlotItem
from qtpy import QtCore
from qtpy.QtWidgets import *


class ImageViewModule(QFrame):
    """
    This class wraps the pyqt imageview model, takes care of configuring it and adds
    a set image method to it
    """

    def __init__(self, main_widget, histogram=True, crop_selector=False):
        super().__init__()

        self.main_widget = main_widget
        # self.setMinimumWidth(600)
        # self.setMinimumHeight(300)
        # self.setStyleSheet("ImageViewModule {margin:5px; border:1px solid rgb(50, 65, "
        #                    "75);} ")
        self.setStyleSheet("ImageViewModule {margin:0px; border:0px  solid rgb(50, 65, "
                           "75); padding: 0px;} ")
        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)

        # self.layout.setAlignment(Qt.AlignHCenter)
        self.image_label = QLabel()
        self.layout.addWidget(self.image_label)
        self.setLayout(self.layout)
        # self.already_loaded = True
        # self.no_image_message = QPushButton("Please open a dataset first")
        # self.no_image_message.clicked.connect(main_widget.open_file_dialog)
        # self.no_image_message.setStyleSheet("QPushButton {font-size:80;}")
        self.plot_item = PlotItem()
        self.image_view = ImageView(view=self.plot_item)
        self.image_view.keyPressEvent = self.keyPressEvent
        self.image_view.ui.layoutWidget.setContentsMargins(0, 0, 0, 0)
        # self.image_view.ui.roiBtn.hide()
        self.image_view.ui.menuBtn.hide()
        if not histogram:
            self.image_view.ui.histogram.hide()
        if not crop_selector:
            self.image_view.ui.roiBtn.hide()
        # self.image_view.getRoiPlot().hide()
        self.image_item = self.image_view.getImageItem()
        self.image_view.installEventFilter(self.main_widget.eventFilterCustom)
        self.image_item.installEventFilter(self.main_widget.eventFilterCustom)
        self.plot_item.installEventFilter(self.main_widget.eventFilterCustom)
        self.layout.addWidget(self.image_view)

    @property
    def data_handler(self):
        return self.main_widget.data_handler
    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key_Space and False:
            if self.image_view.playRate == 0:
                fps = (self.image_view.getProcessedImage().shape[0] - 1) / (
                            self.image_view.tVals[-1] - self.image_view.tVals[0])
                self.image_view.play(fps)
                # print fps
            else:
                self.image_view.play(0)
            ev.accept()
        elif ev.key() == QtCore.Qt.Key_Home:
            self.image_view.setCurrentIndex(0)
            self.image_view.play(0)
            ev.accept()
        elif ev.key() == QtCore.Qt.Key_End:
            self.image_view.setCurrentIndex(
                self.image_view.getProcessedImage().shape[0] - 1)
            self.image_view.play(0)
            ev.accept()
        elif ev.key() in self.image_view.noRepeatKeys:
            ev.accept()
            if ev.isAutoRepeat():
                return
            self.image_view.keysPressed[ev.key()] = 1
            self.image_view.evalKeyState()
        else:
            QWidget.keyPressEvent(self.image_view, ev)
    def setImage(self, data):

        # if self.already_loaded == False:
        #     print("changed image")
        #     self.already_loaded = True
        #     self.layout.removeWidget(self.no_image_message)
        #     self.no_image_message.deleteLater()
        #     # self.layout.setAlignment(Qt.AlignLeft)
        #     self.image_view = ImageView()
        #
        #     self.layout.addWidget(self.image_view)
        self.image_view.setImage(data, levelMode='mono', autoRange=True,
                                 autoLevels=True, autoHistogramRange=True)
        bottom_5 = np.percentile(data, 5)
        top_5 = np.percentile(data, 95)
        bottom_2 = np.percentile(data, 2)
        top_2 = np.percentile(data, 98)
        self.image_view.setLevels(bottom_5, top_5)
        self.image_view.setHistogramRange(bottom_2, top_2)



import logging
import os
from functools import reduce
from typing import List

import dask
from dask import delayed

from cidan.LSSC.SpatialBox import SpatialBox
from cidan.LSSC.functions.data_manipulation import reshape_to_2d_over_time
from cidan.LSSC.functions.eigen import generateEigenVectors, saveEigenVectors, \
    loadEigenVectors, saveEmbedingNormImage, createEmbedingNormImageFromMultiple
from cidan.LSSC.functions.embeddings import calcAffinityMatrix
from cidan.LSSC.functions.progress_bar import printProgressBarROI
from cidan.LSSC.functions.roi_extraction import roi_extract_image, merge_rois
from cidan.LSSC.functions.save_test_images import save_roi_images
from cidan.LSSC.functions.widefield_functions import *
# from explorations.load_mat import load_dataset_widefield_test

logger1 = logging.getLogger("cidan.LSSC.process_data")


def process_data(*, test_images: bool, test_output_dir: str,
                 save_dir: str, save_intermediate_steps: bool,
                 image_data: List[np.ndarray],
                 image_data_filtered: List[np.ndarray],
                 shape: List[np.ndarray],
                 crop: List[np.ndarray],
                 slicing: List[np.ndarray],
                 eigen_vectors_already_generated: bool,
                 save_embedding_images: bool,
                 total_num_time_steps: int, total_num_spatial_boxes: int,
                 spatial_overlap: int, metric: str, knn: int,
                 accuracy: int, connections: int, normalize_w_k: int, num_eig=25,
                 merge: bool,
                 num_rois: int, refinement: bool, num_eigen_vector_select: int,
                 max_iter: int, roi_size_min: int, fill_holes: bool,
                 elbow_threshold_method: bool, elbow_threshold_value: float,
                 eigen_threshold_method: bool,
                 eigen_threshold_value: float, merge_temporal_coef: float,
                 roi_size_max: int, pca: bool, pca_data: np.ndarray,
                 eigen_accuracy: int, roi_eccentricity_limit: float,
                 local_max_method: bool, area_stop_threshold=.95, progress_signal=None,
                 widefield=False,
                 image_data_mask=None):
    logger1.debug("""Inputs: num_threads {0},test_images {1}, test_output_dir {2},
                 save_dir {3}, save_intermediate_steps {4},
                 load_data {5}, data_path {6},
                 image_data {7},
                 eigen_vectors_already_generated {8},
                 save_embedding_images {9},
                 total_num_time_steps {10}, total_num_spatial_boxes {11},
                 spatial_overlap {12}, filter {13}, median_filter {14},
                 median_filter_size {15},
                 z_score {16}, slice_stack {17},
                 slice_every {18}, slice_start {19}, metric {20}, knn {21},
                 accuracy {22}, connections {23}, normalize_w_k {24}, num_eig {25},
                 merge {26},
                 num_rois {27}, refinement {28}, num_eigen_vector_select {29},
                 max_iter {30}, roi_size_min {31}, fill_holes {32},
                 elbow_threshold_method {33}, elbow_threshold_value {34},
                 eigen_threshold_method {35},
                 eigen_threshold_value {36}, merge_temporal_coef {37},
                 roi_size_max {38}""".format("none", test_images, test_output_dir,
                                             save_dir, save_intermediate_steps,
                                             "", "",
                                             image_data_filtered,
                                             eigen_vectors_already_generated,
                                             save_embedding_images,
                                             total_num_time_steps,
                                             total_num_spatial_boxes,
                                             spatial_overlap, "", "", "", "", "", "",
                                             "",
                                             "", metric, knn,
                                             accuracy, connections, normalize_w_k,
                                             num_eig,
                                             merge,
                                             num_rois, refinement,
                                             num_eigen_vector_select,
                                             max_iter, roi_size_min, fill_holes,
                                             elbow_threshold_method,
                                             elbow_threshold_value,
                                             eigen_threshold_method,
                                             eigen_threshold_value, merge_temporal_coef,
                                             roi_size_max))
    # TODO Make after eigen vector make function to save intermediate embeding norm
    #  for each spatial box
    # TODO Rewrite to take in a list of loaded datasets

    # TODO add assertions to make sure input splits work for dataset
    if save_intermediate_steps and not eigen_vectors_already_generated:
        filelist = [f for f in os.listdir(os.path.join(save_dir, "eigen_vectors"))]
        for f in filelist:
            os.remove(os.path.join(os.path.join(save_dir, "eigen_vectors"), f))
    if not os.path.isdir(os.path.join(save_dir, "temp_files/")):
        os.mkdir(os.path.join(save_dir, "temp_files/"))
    if not os.path.isdir(os.path.join(save_dir, "temp_files/embedding")):
        os.mkdir(os.path.join(save_dir, "temp_files/embedding"))
    if not os.path.isdir(os.path.join(save_dir, "temp_files/rois")):
        os.mkdir(os.path.join(save_dir, "temp_files/rois"))
    filelist = [f for f in os.listdir(os.path.join(save_dir, "temp_files/embedding"))]
    for f in filelist:
        os.remove(os.path.join(os.path.join(save_dir, "temp_files/embedding"), f))
    filelist = [f for f in os.listdir(os.path.join(save_dir, "temp_files/rois"))]
    for f in filelist:
        os.remove(os.path.join(os.path.join(save_dir, "temp_files/rois"), f))
    # shape = [image_data_filtered[0].shape[1], image_data_filtered[0].shape[2]]
    logger1.debug("image shape {0}".format(shape))
    # print("Creating {} spatial boxes".format(total_num_spatial_boxes))
    printProgressBarROI(total_num_spatial_boxes=total_num_spatial_boxes,
                        total_num_time_steps=total_num_time_steps,
                        save_dir=save_dir, progress_signal=progress_signal)
    spatial_boxes = [SpatialBox(box_num=x, total_boxes=total_num_spatial_boxes,
                                spatial_overlap=spatial_overlap, image_shape=shape)
                     for x in range(total_num_spatial_boxes)]
    all_rois = []
    all_boxes_eigen_vectors = []
    for spatial_box in spatial_boxes:
        spatial_box_data_list = [convert_to_float(spatial_box.extract_box(x)) for x in
                                 image_data_filtered]
        if pca:
            spatial_box_data_list_pca = [spatial_box.extract_box(x) for x in pca_data]
        if len(spatial_box_data_list) == 1:
            spatial_box_data_list = dask.compute(*spatial_box_data_list)
            if pca:
                spatial_box_data_list_pca = dask.compute(*spatial_box_data_list_pca)


        if total_num_time_steps != 1 and len(spatial_box_data_list) == 1 and not pca:
            time_boxes = [(x * (image_data_filtered[0].shape[0] // total_num_time_steps),
                           (x + 1) * (image_data_filtered[0].shape[0] //
                                      total_num_time_steps))
                          for x in range(total_num_time_steps)]


        all_eigen_vectors_list = []
        if not eigen_vectors_already_generated:
            for temporal_box_num, time_box_data in enumerate(
                    time_boxes if total_num_time_steps != 1 and len(
                        spatial_box_data_list) == 1 else spatial_box_data_list):
                # TODO make sure memory doesn't take more than 2x
                if total_num_time_steps != 1 and len(
                        spatial_box_data_list) == 1 and not pca:
                    time_box_data = sliceData(spatial_box_data_list,
                                              start_end=time_box_data)
                # time_box_data = saveTempImage(time_box_data, save_dir,
                #                               spatial_box.box_num)

                time_box_data_2d = reshape_to_2d_over_time(time_box_data)
                if widefield:
                    time_box_data_2d = delayed(mask_data_2d)(time_box_data_2d,
                                                             image_data_mask.flatten())
                if pca:
                    time_box_data_2d_pca = reshape_to_2d_over_time(
                        spatial_box_data_list_pca[temporal_box_num])
                logger1.debug(
                    "Time box {0}, start {1}, end {2}, time_box shape {3}, 2d shape {4}".format(
                        temporal_box_num, 0, 0, time_box_data.shape,
                        time_box_data_2d.shape))
                k = calcAffinityMatrix(
                    pixel_list=time_box_data_2d if not pca else time_box_data_2d_pca,
                    metric=metric,
                    knn=knn, accuracy=accuracy,
                    connections=connections,
                    normalize_w_k=25,
                    num_threads=8,
                    spatial_box_num=spatial_box.box_num,
                    temporal_box_num=temporal_box_num,
                    total_num_spatial_boxes=total_num_spatial_boxes,
                    total_num_time_steps=total_num_time_steps, save_dir=save_dir,
                    progress_signal=progress_signal)
                eigen_vectors = generateEigenVectors(K=k,
                                                     num_eig=time_box_data_2d_pca.shape[
                                                         1] if pca and False and num_eig >
                                                               pca_data[temporal_box_num].shape[
                                                                   0] else num_eig,
                                                     accuracy=10 ** (
                                                                 -1 * eigen_accuracy)

                                                     )
                if save_intermediate_steps:
                    eigen_vectors = saveEigenVectors(
                        e_vectors=eigen_vectors if not widefield else delayed(
                            mask_to_data_2d)(eigen_vectors, image_data_mask),
                        spatial_box_num=spatial_box.box_num,
                        time_box_num=temporal_box_num,
                        save_dir=save_dir,
                        total=total_num_time_steps * total_num_spatial_boxes)
                    if widefield:
                        eigen_vectors = delayed(mask_data_2d)(eigen_vectors,
                                                              image_data_mask)

                all_eigen_vectors_list.append(eigen_vectors)
                if test_images:
                    pass
                    # delayed(save_eigen_images)(eigen_vectors=eigen_vectors,
                    #                            output_dir=test_output_dir,
                    #                            image_shape=spatial_box_data.shape,
                    #                            box_num=spatial_box.box_num).compute()

        else:
            for temporal_box_num in range(len(image_data_filtered)):
                all_eigen_vectors_list.append(
                    loadEigenVectors(spatial_box_num=spatial_box.box_num,
                                     time_box_num=temporal_box_num,
                                     save_dir=save_dir))
                if widefield:
                    all_eigen_vectors_list[-1] = delayed(mask_data_2d)(
                        all_eigen_vectors_list[-1], image_data_mask)

        all_eigen_vectors = delayed(np.hstack)(all_eigen_vectors_list)
        all_boxes_eigen_vectors.append(all_eigen_vectors)
        if save_embedding_images:
            all_eigen_vectors = saveEmbedingNormImage(
                e_vectors=all_eigen_vectors if not widefield else delayed(
                    mask_to_data_2d)(all_eigen_vectors, image_data_mask),
                image_shape=spatial_box.shape,
                save_dir=save_dir,
                spatial_box_num=spatial_box.box_num)
            if widefield:
                all_eigen_vectors = delayed(mask_data_2d)(all_eigen_vectors,
                                                          image_data_mask)

        data_all = spatial_box.extract_box(stack_crop_slice_data(image_data,slicing, crop))

        rois = roi_extract_image(e_vectors=all_eigen_vectors,
                                 original_shape=data_all.shape,
                                 original_2d_vol=reshape_to_2d_over_time(
                                     data_all),
                                 merge=merge,
                                 num_rois=num_rois, refinement=refinement,
                                 num_eigen_vector_select=num_eigen_vector_select,
                                 max_iter=max_iter,
                                 roi_size_min=roi_size_min,
                                 fill_holes=fill_holes,
                                 elbow_threshold_method=elbow_threshold_method,
                                 elbow_threshold_value=elbow_threshold_value,
                                 eigen_threshold_method=eigen_threshold_method,
                                 eigen_threshold_value=eigen_threshold_value,
                                 merge_temporal_coef=merge_temporal_coef,
                                 roi_size_limit=roi_size_max,
                                 box_num=spatial_box.box_num, print_progress=True,
                                 total_num_time_steps=total_num_time_steps,
                                 total_num_spatial_boxes=total_num_spatial_boxes,
                                 roi_eccentricity_limit=roi_eccentricity_limit,
                                 save_dir=save_dir, local_max_method=local_max_method,
                                 progress_signal=progress_signal, widefield=widefield,
                                 image_data_mask=image_data_mask,
                                 area_stop_threshold=area_stop_threshold)
        if test_images:
            pass
            # delayed(save_roi_images)(
            #     roi_list.json=spatial_box.redefine_spatial_cord_1d(rois),
            #                              output_dir=test_output_dir,
            #                              image_shape=shape,
            #                              box_num=spatial_box.box_num).compute()
        all_rois.append(spatial_box.redefine_spatial_cord_1d(rois))
    all_rois = delayed(reduce)(lambda x, y: x + y, all_rois)
    all_rois = all_rois.compute()
    data_all = stack_crop_slice_data(image_data, slicing, crop)

    all_rois_merged = delayed(merge_rois)(roi_list=all_rois,
                                          temporal_coefficient=merge_temporal_coef,
                                          original_2d_vol=reshape_to_2d_over_time(
                                              data_all), widefield=widefield).compute()

    if test_images:
        delayed(save_roi_images)(roi_list=all_rois_merged,
                                 output_dir=test_output_dir,
                                 image_shape=shape, box_num="all").compute()
    if save_embedding_images and save_intermediate_steps:
        createEmbedingNormImageFromMultiple(spatial_box_list=spatial_boxes,
                                            save_dir=save_dir,
                                            num_time_steps=len(image_data_filtered))

    return all_rois_merged


# if __name__ == '__main__':
#     # process_data(
#     #
#     #              test_images=True,
#     #              save_dir="/Users/sschickler/Code_Devel/LSSC-python/tests/test_files/save_dir",
#     #              save_intermediate_steps=False,
#     #              eigen_vectors_already_generated=False,
#     #              save_embedding_images=False,
#     #              test_output_dir="/Users/sschickler/Code_Devel/LSSC-python/tests/test_files/save_dir",
#     #              image_data_filtered=None,
#     #              total_num_time_steps=4, total_num_spatial_boxes=4, spatial_overlap=10,
#     #
#     #              metric="l2", knn=50, accuracy=59, connections=60,
#     #              num_eig=50, normalize_w_k=2, merge=True,
#     #              num_rois=25, refinement=True,
#     #              num_eigen_vector_select=5,
#     #              max_iter=400, roi_size_min=30,
#     #              fill_holes=True,
#     #              elbow_threshold_method=True,
#     #              elbow_threshold_value=1,
#     #              eigen_threshold_method=True,
#     #              eigen_threshold_value=.5,
#     #              merge_temporal_coef=.01,
#              roi_size_max=600)
# with performance_report(filename="dask-report.html"):
# process_data(num_threads=1, load_data=True,
#              data_path="/Users/sschickler/Code Devel/LSSC-python/input_images/dataset_1",
#              test_images=False,
#              test_output_dir="/Users/sschickler/Documents/LSSC-python/output_images/16",
#              image_data=None,
#              save_dir="/Users/sschickler/Code DEvel/LSSC-python/output_images/16",
#              save_embedding_images=True,
#              save_intermediate_steps=True,
#              eigen_vectors_already_generated=False,
    #              total_num_time_steps=20, total_num_spatial_boxes=16, spatial_overlap=30,
    #              filter=True, median_filter_size=(1, 3, 3), median_filter=True,
    #              z_score=False, slice_stack=False, slice_every=10, slice_start=0,
    #              metric="l2", knn=50, accuracy=59, connections=60,
    #              num_eig=50, normalize_w_k=2, merge=True,
    #              num_rois=25, refinement=True,
    #              num_eigen_vector_select=5,
    #              max_iter=400, roi_size_min=30,
    #              fill_holes=True,
    #              elbow_threshold_method=True,
    #              elbow_threshold_value=1,
    #              eigen_threshold_method=True,
    #              eigen_threshold_value=.5,
    #              merge_temporal_coef=.01,
    #              roi_size_max=600)


@delayed
def sliceData(data, start_end):
    return data[0][start_end[0]:start_end[1], :, :]


@delayed
def convert_to_float(data):
    return data.astype(np.float32)


@delayed
def stack_crop_slice_data(image_data, slicing, crop):
    if crop:
        data_all = np.vstack(
            [x[slicing[0]::slicing[1], crop[0][0]:crop[0][1], crop[1][0]:crop[1][1]] for
             x in image_data])
    else:
        data_all = np.vstack([x[slicing[0]::slicing[1], :, :] for x in image_data])
    return data_all


def main():
    # #### Wide field data part
    data_all, mask = load_dataset_widefield_test()
    data_all = data_all
    data_all = data_all.reshape((mask.shape[0], mask.shape[1], -1)).transpose((2, 0, 1))
    eigen_params = {
        "eigen_vectors_already_generated": False,
        "num_eig": 50,
        "normalize_w_k": 32,
        "metric": "l2",
        "knn": 50,
        "accuracy": 75,
        "eigen_accuracy": 8,
        "connections": 40

    }
    roi_extraction_params = {
        "elbow_threshold_method": True,
        "elbow_threshold_value": .95,
        "eigen_threshold_method": True,
        "eigen_threshold_value": .9,
        "num_eigen_vector_select": 1,
        "merge_temporal_coef": .99,
        "roi_size_min": 30,
        "roi_size_max": 600,
        "merge": True,
        "num_rois": 60,
        "fill_holes": True,
        "refinement": True,
        "max_iter": 100,
        "roi_circ_threshold": 0,
        "roi_eccentricity_limit": .9,
        "local_max_method": False

    }
    process_data(test_images=False, test_output_dir="",
                 save_dir="/Users/sschickler/Code_Devel/Widefield/test/",
                 shape=data_all.shape[1:],
                 save_intermediate_steps=True,
                 image_data_filtered=[data_all],
                 image_data=[data_all],
                 crop=False,
                 slicing=[0, 1],
                 eigen_vectors_already_generated=False,
                 save_embedding_images=True,
                 total_num_time_steps=1,
                 total_num_spatial_boxes=1,
                 spatial_overlap=5,

                 metric=eigen_params["metric"],
                 knn=eigen_params["knn"],
                 accuracy=eigen_params["accuracy"],
                 eigen_accuracy=eigen_params[
                     "eigen_accuracy"],
                 connections=eigen_params[
                     "connections"],
                 normalize_w_k=eigen_params[
                     "normalize_w_k"],
                 num_eig=eigen_params["num_eig"],
                 merge=roi_extraction_params["merge"],
                 num_rois=roi_extraction_params[
                     "num_rois"],
                 refinement=roi_extraction_params[
                     "refinement"],
                 num_eigen_vector_select=
                 roi_extraction_params[
                     "num_eigen_vector_select"],
                 max_iter=roi_extraction_params[
                     "max_iter"],
                 roi_size_min=roi_extraction_params[
                     "roi_size_min"],
                 fill_holes=roi_extraction_params[
                     "fill_holes"],

                 elbow_threshold_method=
                 roi_extraction_params[
                     "elbow_threshold_method"],
                 elbow_threshold_value=
                 roi_extraction_params[
                     "elbow_threshold_value"],
                 eigen_threshold_method=
                 roi_extraction_params[
                     "eigen_threshold_method"],
                 eigen_threshold_value=
                 roi_extraction_params[
                     "eigen_threshold_value"],
                 merge_temporal_coef=
                 roi_extraction_params[
                     "merge_temporal_coef"],
                 roi_size_max=roi_extraction_params[
                     "roi_size_max"],
                 pca=False,
                 pca_data=False,
                 roi_eccentricity_limit=roi_extraction_params[
                     "roi_eccentricity_limit"],
                 local_max_method=
                 roi_extraction_params[
                     "local_max_method"],
                 widefield=True,
                 area_stop_threshold=area_stop_threshold,

                 progress_signal=None, image_data_mask=mask)

    # number_times_run = 10
    # accuracies = [10,20,40,60,100]
    # crops = [[10,245],[60, 245],[120,245]]
    # time_lengths = [100,200,400,600, 800,1200]
    # results = pd.DataFrame.from_dict({"accuracy":[], "num_pixels":[], "time_steps":[], "run_num":[],"time":[]})
    # for x in range(number_times_run):
    #     for crop in crops:
    #         for time_length in time_lengths:
    #             for accuracy in accuracies:
    #                 data_current = data_all[:time_length,crop[0]:crop[1],crop[0]:crop[1]]
    #                 time_start = time.time()
    #                 process_data(test_images=False, test_output_dir="",
    #                              save_dir="/Users/sschickler/Code_Devel/Widefield/test2/",
    #                              shape=data_current.shape[1:],
    #                              save_intermediate_steps=True,
    #                              image_data_filtered=[data_current for _ in range(1)],
    #                              image_data=[data_current for _ in range(1)],
    #                              crop=False,
    #                              slicing=[0, 1],
    #                              eigen_vectors_already_generated=False,
    #                              save_embedding_images=True,
    #                              total_num_time_steps=1,
    #                              total_num_spatial_boxes=1,
    #                              spatial_overlap=5,
    #
    #                              metric=eigen_params["metric"],
    #                              knn=eigen_params["knn"],
    #                              accuracy=accuracy,
    #                              eigen_accuracy=eigen_params[
    #                                  "eigen_accuracy"],
    #                              connections=accuracy,
    #                              normalize_w_k=eigen_params[
    #                                  "normalize_w_k"],
    #                              num_eig=eigen_params["num_eig"],
    #                              merge=roi_extraction_params["merge"],
    #                              num_rois=roi_extraction_params[
    #                                  "num_rois"],
    #                              refinement=roi_extraction_params[
    #                                  "refinement"],
    #                              num_eigen_vector_select=
    #                              roi_extraction_params[
    #                                  "num_eigen_vector_select"],
    #                              max_iter=roi_extraction_params[
    #                                           "max_iter"]+100,
    #                              roi_size_min=roi_extraction_params[
    #                                  "roi_size_min"],
    #                              fill_holes=roi_extraction_params[
    #                                  "fill_holes"],
    #
    #                              elbow_threshold_method=
    #                              roi_extraction_params[
    #                                  "elbow_threshold_method"],
    #                              elbow_threshold_value=
    #                              roi_extraction_params[
    #                                  "elbow_threshold_value"],
    #                              eigen_threshold_method=
    #                              roi_extraction_params[
    #                                  "eigen_threshold_method"],
    #                              eigen_threshold_value=
    #                              roi_extraction_params[
    #                                  "eigen_threshold_value"],
    #                              merge_temporal_coef=
    #                              roi_extraction_params[
    #                                  "merge_temporal_coef"],
    #                              roi_size_max=roi_extraction_params[
    #                                  "roi_size_max"],
    #                              pca=False,
    #                              pca_data=False,
    #                              roi_eccentricity_limit=roi_extraction_params[
    #                                  "roi_eccentricity_limit"],
    #                              local_max_method=
    #                              roi_extraction_params[
    #                                  "local_max_method"],
    #                              widefield=False,
    #
    #                              progress_signal=None,
    #                              # image_data_mask=mask
    #                              )
    #                 time_end = time.time()
    #                 results = results.append({"accuracy":accuracy, "num_pixels":(crop[1]-crop[0])**2, "time_steps":time_length, "run_num":x, "time":time_end-time_start}, ignore_index=True)
    #                 results.to_csv("out.csv")

if __name__ == '__main__':
    main()



import numpy as np
from dask import compute
from scipy.sparse import csr_matrix

from cidan.LSSC.functions.eigen import generateEigenVectors
from cidan.LSSC.functions.embeddings import calcAffinityMatrix

knn= 45
num_elements = 40000
normalize_w_k =30
def test_eigen():

    shape = (200, 200)
    num_points = 15
    shape_2d = (num_points * 25, 2)
    save_dir = "/Users/sschickler/Code Devel/LSSC-python/input_images/test2"
    image = np.zeros(shape_2d, dtype=np.float)
    for i in range(0, num_points, 1):
        for j in range(0, 25, 1):
            image[i * 25 + j, 0] = i
            image[i * 25 + j, 1] = j

    num_eig = 50
    # image_2d = reshape_to_2d_over_time(np.transpose(image,(2,0,1)))
    K = calcAffinityMatrix(pixel_list=image, metric="l2", knn=20,
                           accuracy=80, connections=30,
                           normalize_w_k=15, num_threads=8, spatial_box_num=0,
                           temporal_box_num=0).compute()
    # K_new = np.zeros((15 * 25, 15 * 25))
    # for i in range(0, num_points, 1):
    #     for j in range(0, 25, 1):
    #         if (i > 0):
    #             K_new[i * 25 + j, (i - 1) * 25 + j] = 1
    #         if (i < num_points - 1):
    #             K_new[i * 25 + j, (i + 1) * 25 + j] = 1
    #         if (j > 0):
    #             K_new[i * 25 + j, (i) * 25 + j - 1] = 1
    #         if (j < 25 - 1):
    #             K_new[i * 25 + j, (i) * 25 + j + 1] = 1

    e_vectors = generateEigenVectors(K=csr_matrix(K), num_eig=num_eig,
                                     save_dir="").compute()
    e_vectors = np.array(compute(e_vectors)[0])
    num = 0
    print(np.max(e_vectors[:, num]), np.min(e_vectors[:, num]))
    assert np.isclose(np.min(e_vectors[:, 0]), np.max(e_vectors[:, 0]))
    # plt.pcolormesh(e_vectors.transpose((1,0)))

    # plt.scatter(image.transpose((1, 0))[0], image.transpose((1, 0))[1],
    #             c=e_vectors[:, num])
    # plt.show()
    # save_image(image[:,:,0],0,shape,save_dir)
    # for x in range(num_eig):
    #     save_image(e_vectors[:, x], x + 1, shape, save_dir)




